import os.path
import sys

import rope.base.pyobjects
import rope.base.pynames
import rope.base.builtins

from rope.base import exceptions

from .ropehints import HintProvider, get_attribute_scope_path

def add_django_support(composite_provider, settings='settings'):
    return composite_provider.add_hint_provider(
        DjangoHintProvider(composite_provider.project, settings))

def get_path_and_package(module_path, project_root):
    packages = [os.path.basename(module_path).rpartition('.')[0]]
    while True:
        path = os.path.dirname(module_path)
        if path == module_path:
            break

        module_path = path

        if module_path == project_root:
            break

        if os.path.exists(os.path.join(module_path, '__init__.py')):
            packages.append(os.path.basename(module_path))
        else:
            break

    return module_path, '.'.join(reversed(packages))

loaded_django_modules = {}
def load_django_module(pymodule, project_root):
    path = os.path.realpath(pymodule.resource.real_path)
    try:
        return loaded_django_modules[path]
    except KeyError:
        pass

    syspath, module = get_path_and_package(path, project_root)
    if syspath not in sys.path:
        sys.path.append(syspath)

    __import__(module)

    loaded_django_modules[path] = sys.modules[module]
    return sys.modules[module]


class DjangoHintProvider(HintProvider):
    def __init__(self, project, settings):
        super(DjangoHintProvider, self).__init__(project)

        self.settings = settings

    def get_class_attributes(self, scope_path, pyclass, attrs):
        """:type pyclass: rope.base.pyobjectsdef.PyClass"""

        if not any('django.db.models.base.Model' in get_attribute_scope_path(c)
                for c in pyclass.get_superclasses()):
            return {}

        os.environ['DJANGO_SETTINGS_MODULE'] = self.settings

        module = load_django_module(pyclass.get_module(), self.project.address)
        model = getattr(module, pyclass.get_name())()

        attrs = {}
        for name in model._meta.get_all_field_names():
            f = model._meta.get_field_by_name(name)[0]

            if f.__class__.__name__ == 'ForeignKey':
                related_model_name = f.rel.to.__module__ + '.' + f.rel.to.__name__
                attrs[name] = rope.base.pynames.DefinedName(
                    rope.base.pyobjects.PyObject(self.get_type(related_model_name).get_object()))
                attrs[f.attname] = rope.base.pynames.DefinedName(rope.base.pyobjects.PyObject(None))
            elif f.__class__.__name__ == 'RelatedObject':
                related_model_name = f.model.__module__ + '.' + f.model.__name__
                attrs[name] = DjangoObjectsName(self.get_type(related_model_name).get_object(),
                    self.get_type('django.db.models.manager.Manager'))
            else:
                attrs[name] = rope.base.pynames.DefinedName(rope.base.pyobjects.PyObject(None))

        attrs['objects'] = DjangoObjectsName(pyclass,
            self.get_type('django.db.models.manager.Manager'))

        return attrs


class DjangoObjectsName(rope.base.pynames.PyName):
    def __init__(self, pyclass, name):
        self._orig_name = name
        self.model_type = pyclass

    def get_object(self):
        return DjangoObjectsObject(self.model_type, self._orig_name.get_object())


def proxy(obj):
    cls = obj.__class__
    class Cls(cls):
        def __init__(self, obj):
            self._orig = obj

        def __getattribute__(self, name):
            obj = cls.__getattribute__(self, '_orig')
            print 'getattr', obj, name
            return cls.__getattribute__(obj, name)

    return Cls(obj)

class GetHolder:
    def __init__(self, result):
        self.result = result

    def get(self, *args):
        return self.result

class DjangoObjectsObject(object):
    def __init__(self, pyclass, obj):
        self.model_type = pyclass
        self._orig_object = obj
        self.type = obj.type

    def get_attributes(self):
        attrs = self._orig_object.get_attributes()

        attrs['get'].pyobject.returned = GetHolder(self.model_type)
        attrs['filter'].pyobject.returned = GetHolder(self)
        attrs['exclude'].pyobject.returned = GetHolder(self)
        attrs['extra'].pyobject.returned = GetHolder(self)
        attrs['order_by'].pyobject.returned = GetHolder(self)
        attrs['select_related'].pyobject.returned = GetHolder(self)
        attrs['all'].pyobject.returned = GetHolder(self)

        attrs['__getitem__'] = rope.base.pynames.DefinedName(
            SimpleFunction(self.model_type, ['index']))

        attrs['__iter__'] = rope.base.pynames.DefinedName(
            SimpleFunction(rope.base.builtins.get_iterator(self.model_type)))

        return attrs

    def __getattr__(self, name):
        return getattr(self._orig_object, name)

    def get_attribute(self, name):
        try:
            return self.get_attributes()[name]
        except KeyError:
            raise exceptions.AttributeNotFoundError('Attribute %s not found' % name)

    def __getitem__(self, key):
        return self.get_attribute(key)

    def __contains__(self, key):
        return key in self.get_attributes()


class SimpleFunction(rope.base.pyobjects.AbstractFunction):

    def __init__(self, returned=None, argnames=[]):
        rope.base.pyobjects.AbstractFunction.__init__(self)
        self.argnames = argnames
        self.returned = returned

    def get_returned_object(self, args):
        return self.returned

    def get_param_names(self, special_args=True):
        return self.argnames


import xml.sax.handler
import rope.base.pynames

import gobject

from .ropehints import HintProvider

def add_gtk_support(composite_provider):
    existing_modules = composite_provider.project.prefs['extension_modules']

    for m in ('gtk._gtk', 'gtk.gdk', 'gobject._gobject', 'pango'):
        if m not in existing_modules:
            existing_modules.append(m)

    composite_provider.project.prefs['extension_modules'] = existing_modules

    return composite_provider.add_hint_provider(PyGtkHintProvider(composite_provider.project))


class ResourceAsModule(object):
    def __init__(self, resource):
        self.resource = resource

    def get_resource(self):
        return self.resource

    def get_module(self):
        return self


class GladeName(rope.base.pynames.PyName):
    def __init__(self, pyobject, module, line):
        self.pyobject = pyobject
        self.line = line
        self.module = module

    def get_object(self):
        return self.pyobject

    def get_definition_location(self):
        return self.module, self.line


class PyGtkHintProvider(HintProvider):
    def __init__(self, project):
        super(PyGtkHintProvider, self).__init__(project)
        self.gtk_aware_classes = {}
        self.cache = {}
        self.func_cache = {}
        self.handlers = {}

    def get_pygtk_class_name(self, gtk_class_name):
        return gtk_class_name.replace('Gtk', 'gtk.', 1) + '()'

    def process_glade(self, scope_path, force=False):
        glade_file, processed = self.gtk_aware_classes[scope_path]
        if processed and not force:
            return

        resource = self.project.get_resource(glade_file)

        handler = GladeHandler()
        xml.sax.parseString(open(resource.real_path).read(), handler)

        attrs = {}
        for id, cls, line in handler.objects:
            type = self.get_type(self.get_pygtk_class_name(cls))
            attrs[id] = GladeName(type.get_object(), ResourceAsModule(resource), line)

        self.cache[scope_path] = attrs

        for name, (cls, signal) in handler.signals.iteritems():
            self.handlers.setdefault(scope_path, {})[name] = cls, signal

        self.gtk_aware_classes[scope_path] = glade_file, True

    def get_class_attributes(self, scope_path, pyclass, attrs):
        attrs = {}
        if scope_path in self.gtk_aware_classes:
            self.process_glade(scope_path)
            for k, v in self.cache[scope_path].iteritems():
                attrs[k] = v

        return attrs

    def add_class(self, scope, glade_file):
        self.gtk_aware_classes[scope] = glade_file, False

    def get_function_param_type(self, pyfunc, name):
        """Resolve function's parameters types from doc string

        :type pyfunc: rope.base.pyobjectsdef.PyFunction
        """

        scope_path = self.get_scope_path(pyfunc.get_scope())
        try:
            params = self.func_cache[scope_path]
        except KeyError:
            class_scope = scope_path.rpartition('.')[0]
            if class_scope not in self.gtk_aware_classes:
                params = {}
            else:
                self.process_glade(class_scope)
                params = self.get_params_for_handler(class_scope, pyfunc)

            self.func_cache[scope_path] = params

        return params.get(name, None)

    def get_params_for_handler(self, class_scope, pyfunc):
        """:type pyfunc: rope.base.pyobjectsdef.PyFunction"""
        try:
            cls, signal = self.handlers[class_scope][pyfunc.get_name()]
        except KeyError:
            return {}

        attrs = {}

        names = pyfunc.get_param_names(False)
        if pyfunc.get_kind() in ('method', 'classmethod'):
            names = names[1:]

        if names:
            attrs[names[0]] = self.get_type(self.get_pygtk_class_name(cls)).get_object()
            names = names[1:]

        if names:
            for i, t in enumerate(gobject.signal_query(signal, str(cls))[-1]):
                try:
                    tname = self.get_type(self.get_pygtk_class_name(t.name))
                    if tname:
                        attrs[names[i]] = tname.get_object()
                    else:
                        attrs[names[i]] = None
                except IndexError:
                    break

        return attrs


class GladeHandler(xml.sax.handler.ContentHandler):
    def __init__(self):
        self.objects = []
        self.signals = {}
        self.current_object_class = None

    def startElement(self, name, attrs):
        if name == 'object':
            self.objects.append((attrs['id'], attrs['class'], self._locator.getLineNumber()))
            self.current_object_class = attrs['class']
        elif name == 'signal':
            if self.current_object_class:
                self.signals[attrs['handler']] = self.current_object_class, attrs['name']

    def endElement(self, name):
        if name == 'object':
            self.current_object_class = None


dialog = None
        
def init(manager):
    manager.add_shortcut('quick-open', '<ctrl><alt>r', 'File', "Shows quick open dialog", activate)

def editor_opened(editor):
    import settings
    root = editor.project_root
    if root and root not in settings.recent_projects:            
        settings.recent_projects.append(root)                
    
def activate(editor):
    global dialog
    if not dialog:
        from gui import QuickOpenDialog
        dialog = QuickOpenDialog()

    dialog.show(editor)
        
def quit():
    global dialog
    dialog = dialog



import re

def remove_trailing_spaces(buffer):
    matcher = re.compile(r'(?um)([ \t]+)$')
    while True:
        match = matcher.search(buffer.get_text(*buffer.get_bounds()).decode('utf-8'))
        if match:
            buffer.delete(*map(buffer.get_iter_at_offset, match.span(1)))
        else:
            break



import sys
import traceback

from snaked.core.shortcuts import Shortcut, register_shortcut, get_path_by_key
from snaked.core.prefs import ListSettings

default_enabled_plugins = ['save_positions', 'edit_and_select',
    'python', 'complete_words', 'hash_comment', 'python_flakes', 'goto_line',
    'goto_dir', 'search']

def get_package(name):
    try:
        return sys.modules[name]
    except KeyError:
        __import__(name)
        return sys.modules[name]

def get_plugin(plugin):
    package_name = 'snaked.plugins.' + plugin
    return get_package(package_name)


class ShortcutsHolder(object):
    def __init__(self):
        self.shortcuts = []

    def add_shortcut(self, name, accel, category, desc, callback):
        self.shortcuts.append((name, accel, category, desc, callback))

    def add_editor_preferences(self, on_dialog_created, on_pref_refresh, default_values):
        import snaked.core.prefs
        import snaked.core.gui.editor_prefs

        for k, v in default_values.iteritems():
            snaked.core.prefs.default_prefs.setdefault(k, {}).update(v)

        snaked.core.gui.editor_prefs.on_dialog_created_hooks.append(on_dialog_created)
        snaked.core.gui.editor_prefs.on_pref_refresh_hooks.append(on_pref_refresh)


class PluginManager(object):
    def __init__(self):
        self.restore_enabled_plugins()

        self.loaded_plugins = {}
        self.core_plugins = []

        self.plugin_by_path = {}
        self.shortcuts_by_plugins = {}
        self.binded_shortcuts = {}

    def get_plugin(self, name):
        try:
            return self.loaded_plugins[name]
        except KeyError:
            plugin = get_plugin(name)
            self.init_plugin(plugin, name)
            return plugin

    def init_plugin(self, plugin, name=None):
        name = name or plugin.__name__
        if hasattr(plugin, 'init'):
            holder = ShortcutsHolder()
            plugin.init(holder)
            self.add_shortcuts(plugin, holder)
        self.loaded_plugins[name] = plugin

    def load_core_plugin(self, plugin):
        self.init_plugin(plugin)
        self.core_plugins.append(plugin.__name__)

    def add_shortcuts(self, plugin, holder):
        for name, accel, category, desc, callback in holder.shortcuts:
            shortcut = Shortcut(name, accel, category, desc)
            shortcut.callback = callback

            path = register_shortcut(name, accel, category, desc)

            self.plugin_by_path[path] = (plugin, shortcut)
            self.shortcuts_by_plugins.setdefault(plugin, []).append(shortcut)

    def bind_shortcuts(self, activator, editor):
        for p in self.plugins_for(editor):
            for s in self.shortcuts_by_plugins.get(p, []):
                try:
                    self.binded_shortcuts[activator][s.name]
                except KeyError:
                    self.binded_shortcuts.setdefault(activator, {})[s.name] = True
                    activator.bind_to_name(s.name, self.activate_plugin_shortcut)

    def activate_plugin_shortcut(self, key, modifier, editor, *args):
        try:
            plugin, shortcut = self.plugin_by_path[get_path_by_key(key, modifier)]
        except KeyError:
            return False

        if self.plugin_is_for_editor(plugin, editor):
            shortcut.callback(editor, *args)
            return True

        return False
    activate_plugin_shortcut.provide_key = True

    def plugin_is_for_editor(self, plugin, editor):
        return not hasattr(plugin, 'langs') or editor.lang in plugin.langs

    def plugins_for(self, editor):
        for name in self.enabled_plugins + self.core_plugins:
            try:
                plugin = self.get_plugin(name)
            except Exception:
                editor.message("Can't load %s plugin" % name, 5000)
                self.enabled_plugins.remove(name)
                traceback.print_exc()
                continue

            if self.plugin_is_for_editor(plugin, editor):
                yield plugin

    def editor_opened(self, editor):
        for p in self.plugins_for(editor):
            if hasattr(p, 'editor_opened'):
                try:
                    p.editor_opened(editor)
                except:
                    traceback.print_exc()

    def editor_closed(self, editor):
        for p in self.plugins_for(editor):
            if hasattr(p, 'editor_closed'):
                try:
                    p.editor_closed(editor)
                except:
                    traceback.print_exc()

    def editor_created(self, editor):
        for p in self.plugins_for(editor):
            if hasattr(p, 'editor_created'):
                try:
                    p.editor_created(editor)
                except:
                    traceback.print_exc()

    def quit(self):
        for p in self.loaded_plugins.values():
            if hasattr(p, 'quit'):
                try:
                    p.quit()
                except:
                    traceback.print_exc()

    @property
    def prefs(self):
        return ListSettings('enabled-plugins.db')

    def save_enabled_plugins(self):
        self.prefs.store(self.enabled_plugins)

    def restore_enabled_plugins(self):
        prefs = self.prefs
        if prefs.exists():
            self.enabled_plugins = prefs.load()
        else:
            self.enabled_plugins = default_enabled_plugins

    def unload_unnecessary_plugins(self):
        bad_plugin_names = [name for name in self.loaded_plugins
            if name not in self.enabled_plugins and name not in self.core_plugins]

        for name in bad_plugin_names:
            p = self.loaded_plugins[name]
            if hasattr(p, 'quit'):
                try:
                    p.quit()
                except:
                    traceback.print_exc()

            del self.loaded_plugins[name]
            self.delete_plugin_shortcuts(p)

    def delete_plugin_shortcuts(self, plugin):
        try:
            del self.shortcuts_by_plugins[plugin]
        except KeyError:
            pass

        for path in [path for path, (p,_) in self.plugin_by_path.iteritems() if p is plugin]:
            del self.plugin_by_path[path]

    def show_plugins_prefs(self, editor):
        from snaked.core.gui.plugin_prefs import PluginDialog
        dialog = PluginDialog()
        editor.request_transient_for.emit(dialog.window)

        def set_plugin_list(plugin_list):
            self.enabled_plugins = plugin_list
            self.save_enabled_plugins()
            editor.message('Enabled plugins list saved')
            self.unload_unnecessary_plugins()
            editor.plugins_changed.emit()

        dialog.show(self.enabled_plugins, set_plugin_list)



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'External tools'
desc = 'Allows one to define own commands'

import weakref
import gtk

from snaked.util import idle
from snaked.core.prefs import register_dialog

def init(manager):
    manager.add_shortcut('external-tools', '<alt>x', 'Tools', 'Run tool', run_tool)
    register_dialog('External tools', show_preferences, 'run', 'external', 'tool', 'command')

def to_str(data, encoding='utf-8'):
    if isinstance(data, unicode):
        return data.encode(encoding)

    return data

def get_run_menu(prefs, editor):
    menu = gtk.Menu()
    menu.set_reserve_toggle_size(False)

    has_selection = editor.buffer.get_has_selection()

    any_items = False
    for tool in sorted(prefs):
        if prefs[tool]['stdin'] == 'selection' and not has_selection:
            continue

        if prefs[tool]['langs'].strip() and editor.lang not in map(str.strip,
                to_str(prefs[tool]['langs'].split(','))):
            continue

        any_items = True
        item = gtk.MenuItem(None, True)
        label = gtk.Label()
        label.set_alignment(0, 0.5)
        label.set_markup_with_mnemonic(prefs[tool]['name'])
        if len(tool) < 10:
            label.set_width_chars(10)

        item.add(label)
        menu.append(item)
        item.connect('activate', on_item_activate, weakref.ref(editor), tool, prefs[tool])

    if any_items:
        menu.show_all()
        return menu
    else:
        menu.destroy()
        return None

def run_tool(editor):
    from snaked.core.prefs import load_json_settings

    prefs = load_json_settings('external-tools.conf', {})
    if not prefs:
        editor.message('There is no any tool to run')
        return

    def get_coords(menu):
        win = editor.view.get_window(gtk.TEXT_WINDOW_TEXT)
        x, y, w, h, _ = win.get_geometry()
        x, y = win.get_origin()
        mw, mh = menu.size_request()
        return x + w - mw, y + h - mh, False

    menu = get_run_menu(prefs, editor)
    if not menu:
        editor.message('There is no any tool to run')
        return

    menu.popup(None, None, get_coords, 0, gtk.get_current_event_time())

def on_item_activate(item, editor, name, prefs):
    idle(run, editor(), name, prefs)
    idle(item.get_parent().destroy)

def show_preferences(editor):
    from prefs import PreferencesDialog
    PreferencesDialog().show(editor)

def get_stdin(editor, id):
    if id == 'none':
        return None
    elif id == 'buffer':
        return editor.text
    elif id == 'selection':
        return editor.buffer.get_text(*editor.buffer.get_selection_bounds())
    elif id == 'buffer-or-selection':
        if editor.buffer.get_has_selection():
            return editor.buffer.get_text(*editor.buffer.get_selection_bounds())
        else:
            return editor.text

def replace(editor, bounds, text):
    line = editor.cursor.get_line()
    editor.view.window.freeze_updates()

    editor.buffer.begin_user_action()
    editor.buffer.delete(*bounds)
    editor.buffer.insert_at_cursor(text)
    editor.buffer.end_user_action()
    editor.goto_line(line + 1)
    editor.view.window.thaw_updates()

def insert(editor, iter, text):
    editor.buffer.begin_user_action()
    editor.buffer.insert(iter, text)
    editor.buffer.end_user_action()

def process_stdout(editor, stdout, id):
    if id == 'show-feedback':
        editor.message(stdout, 3000)
    elif id == 'replace-selection':
        replace(editor, editor.buffer.get_selection_bounds(), stdout)
    elif id == 'replace-buffer':
        replace(editor, editor.buffer.get_bounds(), stdout)
    elif id == 'replace-selection-or-buffer':
        if editor.buffer.get_has_selection():
            replace(editor, editor.buffer.get_selection_bounds(), stdout)
        else:
            replace(editor, editor.buffer.get_bounds(), stdout)
    elif id == 'insert':
        insert(editor, editor.cursor, stdout)
    elif id == 'insert-at-end':
        last_line = editor.buffer.get_line_count()
        insert(editor, editor.buffer.get_bounds()[1], stdout)
        editor.goto_line(last_line)
    elif id == 'clipboard':
        clipboard = editor.view.get_clipboard(gtk.gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(stdout)
        editor.message('Command output was placed on clipboard')
    else:
        editor.message('Unknown stdout action ' + id)

def run(editor, name, prefs):
    import shlex
    import os.path
    from subprocess import Popen, PIPE

    command = shlex.split(to_str(prefs['command']))
    if not command:
        editor.message('Tool must define command to run')
        return

    editor.message('Running ' + name)

    stdin = get_stdin(editor, prefs['stdin'])

    current_file = editor.uri
    current_dir = os.path.dirname(current_file)
    current_project = editor.project_root

    command_to_run = ['/usr/bin/env']
    for c in command:
        command_to_run.append(c.replace('%f', current_file).replace(
            '%d', current_dir).replace('%p', current_project))

    stdout, stderr = Popen(command_to_run, stdout=PIPE, stderr=PIPE,
        stdin=PIPE if stdin else None).communicate(stdin)

    if stderr:
        editor.message(stderr, 5000)

    process_stdout(editor, stdout, prefs['stdout'])


"""
This is ropehints module which configures Snaked's python type hinting system.

Enable Django support::

    from snaked.plugins.python.djangohints import add_django_support
    add_django_support(provider)

Enable PyGtk support::

    from snaked.plugins.python.pygtkhints import add_gtk_support
    pygtk = add_gtk_support(provider)

For detailed description look at http://packages.python.org/snaked/python.html#type-hints
"""

def init(provider):
    pass


import os.path
import os
import weakref

import gtk

from snaked.util import (idle, join_to_file_dir, BuilderAware, open_mime, refresh_gui,
    set_activate_the_one_item)
from snaked.core.shortcuts import ShortcutActivator
from snaked.core.prefs import ListSettings, load_json_settings, save_json_settings

import settings
import searcher

class QuickOpenDialog(BuilderAware):
    def __init__(self):
        super(QuickOpenDialog, self).__init__(join_to_file_dir(__file__, 'gui.glade'))
        self.shortcuts = ShortcutActivator(self.window)
        self.shortcuts.bind('Escape', self.escape)
        self.shortcuts.bind('<alt>Up', self.project_up)
        self.shortcuts.bind('<alt>Down', self.project_down)
        self.shortcuts.bind('<ctrl>Return', self.open_mime)
        self.shortcuts.bind('<alt>s', self.focus_search)
        self.shortcuts.bind('<ctrl>o', self.free_open)
        self.shortcuts.bind('<ctrl>p', self.popup_projects)
        self.shortcuts.bind('<ctrl>Delete', self.delete_project)
        self.shortcuts.bind('BackSpace', self.browse_top)

        set_activate_the_one_item(self.search_entry, self.filelist_tree)        
        
    def get_stored_recent_projects(self):
        if self.editor().session:
            return load_json_settings(
                '%s.session' % self.editor().session, {}).get('recent_projects', [])
        else:
            return ListSettings('project-roots.db').load()
            
    def store_recent_projects(self, projects):
        if self.editor().session:
            name = '%s.session' % self.editor().session
            session_settings = load_json_settings(name, {})
            session_settings['recent_projects'] = list(projects)
            save_json_settings(name, session_settings)
        else:
            return ListSettings('project-roots.db').store(projects)
    
    def show(self, editor):
        self.editor = weakref.ref(editor)
        self.update_recent_projects()
        self.update_projects(editor.project_root)
        editor.request_transient_for.emit(self.window)
        
        self.search_entry.grab_focus()
        
        self.window.present()
    
    def update_recent_projects(self):
        saved_projects = self.get_stored_recent_projects()
                
        if any(p not in saved_projects for p in settings.recent_projects):
            [saved_projects.append(p) for p in settings.recent_projects
                if p not in saved_projects]
            self.store_recent_projects(saved_projects)
            settings.recent_projects = saved_projects
            return
            
        if any(p not in settings.recent_projects for p in saved_projects):
            [settings.recent_projects.append(p) for p in saved_projects
                if p not in settings.recent_projects]
    
    def update_projects(self, root):
        old_root = self.get_current_root()
        
        self.projects_cbox.handler_block_by_func(self.on_projects_cbox_changed)
        
        self.projects_cbox.set_model(None)
        self.projectlist.clear()
        
        index = 0
        for i, r in enumerate(settings.recent_projects):
            if r == root:
                index = i
            self.projectlist.append((r,))

        if not len(self.projectlist):
            self.projectlist.append((os.getcwd(),))
        
        self.projects_cbox.set_model(self.projectlist)
        self.projects_cbox.set_active(index)
        
        self.projects_cbox.handler_unblock_by_func(self.on_projects_cbox_changed)

        if self.get_current_root() != old_root:
            self.on_search_entry_changed()
        
    def hide(self):
        self.current_search = None
        self.window.hide()
        
    def on_delete_event(self, *args):
        self.escape()
        return True
    
    def project_up(self):
        idx = self.projects_cbox.get_active()
        idx = ( idx - 1 ) % len(self.projectlist)
        self.projects_cbox.set_active(idx)
        
    def project_down(self):
        idx = self.projects_cbox.get_active()
        idx = ( idx + 1 ) % len(self.projectlist)
        self.projects_cbox.set_active(idx)

    def get_current_root(self):
        try:
            idx = self.projects_cbox.get_active()
            if idx >=0:
                return self.projectlist[idx][0]
        except IndexError:
            pass
            
        return None
    
    def fill_filelist(self, search, current_search):
        self.filelist.clear()
        
        already_matched = {}
        counter = [-1]
        
        def tick():
            counter[0] += 1
            if counter[0] % 50 == 0:
                refresh_gui()
                if self.current_search is not current_search:
                    raise StopIteration()
        
        for m in (searcher.name_start_match, searcher.name_match,
                searcher.path_match, searcher.fuzzy_match):
            for p in searcher.search(self.get_current_root(), '', m(search), already_matched, tick):
                if self.current_search is not current_search:
                    return
                
                already_matched[p] = True            
                self.filelist.append(p)
                
                if len(self.filelist) > 150:
                    self.filelist_tree.columns_autosize()
                    return            

        self.filelist_tree.columns_autosize()

    def fill_with_dirs(self, top='', place=False):
        self.filelist.clear()

        dirs = []
        files = []
        
        if top and not top.endswith('/'):
            top += '/'
        
        root = os.path.join(self.get_current_root(), top)
        for name in os.listdir(root):
            path = os.path.join(root, name)
            if os.path.islink(path):
                path = os.path.readlink(path)
                           
            if os.path.isdir(path):
                dirs.append(name+'/')
            else:
                files.append(name)
        
        place_idx = 0
        for i, name in enumerate(sorted(dirs)):
            if name == place:
                place_idx = i
            self.filelist.append((name, top))

        for name in sorted(files):
            self.filelist.append((name, top))
        
        self.filelist_tree.columns_autosize()
        
        if place and len(self.filelist):
            self.filelist_tree.set_cursor((place_idx,))

    def on_search_entry_changed(self, *args):
        search = self.search_entry.get_text().strip()
        self.current_search = object()
        if search:
            idle(self.fill_filelist, search, self.current_search)
        else:
            idle(self.fill_with_dirs)
                    
    def on_projects_cbox_changed(self, *args):
        self.on_search_entry_changed()

    def get_selected_file(self):
        (model, iter) = self.filelist_tree.get_selection().get_selected()
        if iter:
            name, top = self.filelist.get(iter, 0, 1)
            return os.path.join(self.get_current_root(), top, name), name, top
        else:
            return None, None, None
    
    def open_file(self, *args):
        fname, name, top = self.get_selected_file()
        if fname:
            if os.path.isdir(fname):
                idle(self.fill_with_dirs, os.path.join(top, name), True)
            else:
                self.hide()
                refresh_gui()
                self.editor().open_file(fname)
        
    def open_mime(self):
        fname, name, top = self.get_selected_file()
        if fname:
            self.hide()
            refresh_gui()
            open_mime(fname)

    def focus_search(self):
        self.search_entry.grab_focus()

    def escape(self):
        if hasattr(self.editor(), 'on_dialog_escape'):
            idle(self.editor().on_dialog_escape, self)
        self.hide()
        
    def free_open(self):
        dialog = gtk.FileChooserDialog("Open file...",
            None,
            gtk.FILE_CHOOSER_ACTION_OPEN,
            (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
            gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        
        dialog.set_default_response(gtk.RESPONSE_OK)
        
        response = dialog.run()
        if response == gtk.RESPONSE_OK:
            idle(self.editor().open_file, dialog.get_filename())
            idle(self.hide)
        
        dialog.destroy()

    def popup_projects(self):
        self.projects_cbox.popup()
        
    def delete_project(self):
        if len(self.projectlist):
            current_root = self.get_current_root()
            if current_root == self.editor().project_root:
                self.editor().message('You can not remove current project')
                return
            settings.recent_projects.remove(current_root)
            self.store_recent_projects(settings.recent_projects)
        
            idx = self.projects_cbox.get_active()
            self.projectlist.remove(self.projects_cbox.get_active_iter())
            self.projects_cbox.set_active(idx % len(self.projectlist))
            self.editor().message('Project removed')

    def browse_top(self):
        if not self.filelist_tree.is_focus():
            return False
        
        if self.search_entry.get_text():
            self.editor().message('You are not in browse mode')
            return
            
        fname, name, top = self.get_selected_file()
        if fname:
            if not top:
                self.editor().message('No way!')
            else:
                place = os.path.basename(os.path.dirname(top)) + '/'
                idle(self.fill_with_dirs, os.path.dirname(os.path.dirname(top)), place)


import weakref

from snaked.util import BuilderAware, join_to_file_dir, idle, set_activate_the_one_item

import snaked.core.prefs as prefs

class PreferencesDialog(BuilderAware):
    def __init__(self):
        BuilderAware.__init__(self, join_to_file_dir(__file__, 'prefs.glade'))

        from snaked.core.shortcuts import ShortcutActivator
        self.activator = ShortcutActivator(self.window)
        self.activator.bind('Escape', self.hide)
        self.activator.bind('<alt>s', self.focus_search)

        set_activate_the_one_item(self.search_entry, self.dialogs_view)

    def hide(self):
        self.window.destroy()

    def show(self, editor):
        self.editor = weakref.ref(editor)
        self.fill_dialogs(None)
        editor.request_transient_for.emit(self.window)
        self.window.show()

    def fill_dialogs(self, search):
        self.dialogs.clear()

        for name in sorted(prefs.registered_dialogs):
            keywords, show_func = prefs.registered_dialogs[name]
            if not search or any(w.startswith(search) for w in keywords):
                markup = '<b>%s</b>\n<small>%s</small>' % (
                    name, u' \u2022 '.join(keywords))
                self.dialogs.append((name, markup))

    def on_delete_event(self, *args):
        return False

    def on_search_entry_changed(self, *args):
        search = self.search_entry.get_text().strip().lower()
        idle(self.fill_dialogs, search)

    def activate(self, *args):
        (model, iter) = self.dialogs_view.get_selection().get_selected()
        if iter:
            name = model.get_value(iter, 0)
            prefs.registered_dialogs[name][1](self.editor())
            idle(self.hide)
        else:
            self.editor().message('You need select item')

    def focus_search(self):
        self.search_entry.grab_focus()


import weakref
import re

from snaked.util import idle, join_to_file_dir, BuilderAware, refresh_gui, set_activate_the_one_item
from snaked.core.shortcuts import ShortcutActivator

matcher = re.compile(r'(?m)^(?P<level>[ \t]*)(?P<type>def|class)\s+(?P<name>\w+)\s*(\(|:)')

def get_outline(source):
    last_start = 0
    last_line = 1
    cpath = ()
    cpath_levels = (0,)
    cname = None
    for match in matcher.finditer(source):
        last_line = last_line + source.count('\n', last_start, match.start())
        last_start = match.start()

        level, name = match.group('level', 'name')
        level = len(level)
        
        if level < cpath_levels[-1]:
            while level < cpath_levels[-1]:
                cpath = cpath[:-1]
                cpath_levels = cpath_levels[:-1]
        elif level > cpath_levels[-1]:
            cpath += (cname,)
            cpath_levels += (level, )

        yield cpath, name, last_line
        
        cname = name
            
class OutlineDialog(BuilderAware):
    def __init__(self):
        super(OutlineDialog, self).__init__(join_to_file_dir(__file__, 'outline.glade'))
        self.shortcuts = ShortcutActivator(self.window)
        self.shortcuts.bind('Escape', self.hide)
        self.shortcuts.bind('<alt>s', self.focus_search)
        
        set_activate_the_one_item(self.search_entry, self.outline_tree)

    def show(self, editor):
        self.editor = weakref.ref(editor)
        self.search_entry.grab_focus()
        
        editor.request_transient_for.emit(self.window)
        self.window.present()
        
        idle(self.fill)

    def hide(self):
        self.window.hide()
        
    def on_delete_event(self, *args):
        self.hide()
        return True

    def goto_name(self, *args):
        (model, iter) = self.outline_tree.get_selection().get_selected()
        if iter:
            self.hide()
            self.editor().add_spot()
            self.editor().goto_line(model.get_value(iter, 2))
        else:
            self.editor().message('You need select item')
            
    def on_search_entry_changed(self, *args):
        what = self.search_entry.get_text().strip()
        if what:
            idle(self.filter, what)
        else:
            idle(self.fill)
    
    def fill(self):
        self.outline.clear()
        current_search = object()
        self.current_search = current_search
        
        roots = (None, None)
        ptop = ()
        
        i = 0
        for top, name, line in get_outline(self.editor().text):
            if self.current_search is not current_search:
                return

            if len(top) == len(ptop):
                roots = roots[:-1] + (self.outline.append(roots[-2], (name, '', line)),)
            elif len(top) > len(ptop):
                roots = roots + (self.outline.append(roots[-1], (name, '', line)),)
            else:
                delta = len(ptop) - len(top) + 1
                roots = roots[:-delta] + (self.outline.append(roots[-delta-1], (name, '', line)),)

            ptop = top
            
            if i % 10 == 0:
                self.outline_tree.expand_all()
                self.outline_tree.columns_autosize()
                refresh_gui()
                
            i += 1

            self.outline_tree.expand_all()
            self.outline_tree.columns_autosize()
                
    def filter(self, search):
        self.outline.clear()
        
        current_search = object()
        self.current_search = current_search
        
        already_matched = {}
        i = 0
        
        def name_starts(name):
            return name.startswith(search)
            
        def name_contains(name):
            return search in name
        
        outline = list(get_outline(self.editor().text))
        
        for m in (name_starts, name_contains):
            for top, name, line in outline:
                if self.current_search is not current_search:
                    return
                
                if (top, name) in already_matched: continue
                
                if m(name):
                    already_matched[(top, name)] = True            
                    self.outline.append(None, (name, u'/'.join(top), line))
                
                if i % 10 == 0:
                    refresh_gui()
                    
                i += 1

        self.outline_tree.columns_autosize()

    def focus_search(self):
        self.search_entry.grab_focus()


author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Spell check'
desc = 'Attaches spell checker to current editor'

import weakref
import glib
import gtk

attached_spells = weakref.WeakKeyDictionary()

def init(manager):
    manager.add_shortcut('toggle-spell-check', 'F7', 'Edit', 'Toggle spell check', toggle_spell)

    manager.add_editor_preferences(
        on_preferences_dialog_created,
        on_preferences_dialog_refresh, {
        'default':{
            'spell-check': False
        }
    })

def editor_opened(editor):
    if editor.prefs['spell-check']:
        toggle_spell(editor)

def on_preferences_dialog_created(dialog):
    """:type dialog: snaked.core.gui.editor_prefs.PreferencesDialog"""
    dialog.spell_check = gtk.CheckButton('Spell chec_k')
    dialog.spell_check.connect('toggled', dialog.on_checkbox_toggled, 'spell-check')
    dialog.spell_check.show()

    dialog.vbox.pack_start(dialog.spell_check, False, False)

def on_preferences_dialog_refresh(dialog, pref):
    """:type dialog: snaked.core.gui.editor_prefs.PreferencesDialog"""
    dialog.spell_check.set_active(pref['spell-check'])

def toggle_spell(editor):
    if editor in attached_spells:
        spell = attached_spells[editor]
        spell.detach()
        del attached_spells[editor]
    else:
        try:
            from gtkspell import Spell
            from locale import getdefaultlocale
            attached_spells[editor] = Spell(editor.view, getdefaultlocale()[0])
        except ImportError:
            editor.message('Spellcheck not available. You need to install pygtkspell')
        except glib.GError:
            editor.message('Spellcheck not available. Perhaps you have no dictionaries')



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Snippets'
desc = "SnipMate's clone"

import os.path
import re

import weakref
import gtk, gobject

from gtksourceview2 import CompletionProvider, CompletionProposal
from gtksourceview2 import COMPLETION_ACTIVATION_USER_REQUESTED

from snaked.util import idle
from .parser import parse_snippets_from

loaded_snippets = {}
existing_snippet_contexts = {'not_initialized':True}
snippets_match_hash = {}
completion_providers = {}

stop_managers = weakref.WeakKeyDictionary()

def editor_opened(editor):
    if 'not_initialized' in existing_snippet_contexts:
        existing_snippet_contexts.clear()
        discover_snippet_contexts()

    if not any(ctx in existing_snippet_contexts for ctx in editor.contexts):
        return

    contexts = [c for c in editor.contexts if c in existing_snippet_contexts]
    for ctx in contexts:
        if ctx not in loaded_snippets:
            load_snippets_for(ctx)

        editor.view.get_completion().add_provider(completion_providers[ctx])

    editor.view.connect('key-press-event', on_view_key_press_event,
        contexts, weakref.ref(editor))

    editor.buffer.connect_after('changed', on_buffer_changed)

def load_snippets_for(ctx):
    snippets = parse_snippets_from(existing_snippet_contexts[ctx])
    loaded_snippets[ctx] = snippets
    snippet_names = [s.snippet for s in snippets.values()]
    for name in snippet_names:
        snippets_match_hash.setdefault(ctx, {}).setdefault(len(name), {})[name] = True

    completion_providers[ctx] = SnippetsCompletionProvider(ctx)

def discover_snippet_contexts():
    dirs_to_scan = [os.path.join(os.path.dirname(__file__), 'snippets'),
        os.path.join(os.getenv('XDG_CONFIG_HOME', os.path.expanduser('~/.config')),
            'snaked', 'snippets')
    ]

    for d in dirs_to_scan:
        if os.path.exists(d):
            for name in os.listdir(d):
                path = os.path.join(d, name)
                nm, ext = os.path.splitext(name)
                if ext == '.snippets' and os.path.isfile(path):
                    existing_snippet_contexts[nm] = path

def get_match(iter, ctx):
    names = snippets_match_hash.get(ctx, {})
    if not names:
        return None

    for cnt in sorted(names, reverse=True):
        end = iter.copy()
        end.backward_chars(cnt)

        match = end.get_slice(iter)
        if match in names[cnt]:
            return match

    return None

def get_iter_at_cursor(buffer):
    return buffer.get_iter_at_mark(buffer.get_insert())

def on_view_key_press_event(view, event, contexts, editor_ref):
    if event.keyval == gtk.keysyms.Tab:
        buffer = view.get_buffer()
        cursor = get_iter_at_cursor(buffer)

        matches = {}
        for ctx in contexts:
            match = get_match(cursor, ctx)
            if match:
                matches[ctx] = find_all_snippets(ctx, match)

        if matches:
            return expand_snippet(editor_ref, matches)

        if buffer in stop_managers:
            sm = stop_managers[buffer]
            if sm.cursor_in_snippet_range(cursor):
                return sm.goto_next_stop()
            else:
                del stop_managers[buffer]

    elif event.keyval == gtk.keysyms.ISO_Left_Tab:
        buffer = view.get_buffer()
        cursor = get_iter_at_cursor(buffer)

        if buffer in stop_managers:
            sm = stop_managers[buffer]
            if sm.cursor_in_snippet_range(cursor):
                return sm.goto_next_stop(True)
            else:
                del stop_managers[buffer]

    return False

def on_buffer_changed(buffer):
    if buffer in stop_managers:
        cursor = get_iter_at_cursor(buffer)
        sm = stop_managers[buffer]
        if sm.cursor_in_snippet_range(cursor):
            if sm.snippet_collapsed():
                del stop_managers[buffer]
            else:
                idle(sm.replace_inserts)
        else:
            del stop_managers[buffer]

def find_all_snippets(ctx, match):
    return [s for s in loaded_snippets[ctx].values() if s.snippet == match]

def expand_snippet(editor_ref, matches):
    if not matches:
        return False
    elif len(matches) == 1:
        snippets = matches.values()[0]
        if not snippets:
            return False

        if len(snippets) == 1:
            insert_snippet(editor_ref, editor_ref().cursor, snippets[0])
            return True

    show_proposals(editor_ref().view, editor_ref().cursor, matches.keys())
    return True

match_ws = re.compile(u'(?u)^[ \t]*')
def get_whitespace(start):
    match = match_ws.search(line_text(start))
    if match:
        return match.group(0)
    else:
        return u''

def line_text(iter):
    if not iter.starts_line():
        iter = iter.copy()
        iter.set_line(iter.get_line())

    end = iter.copy()
    if not end.ends_line():
        end.forward_to_line_end()

    return iter.get_text(end)

def insert_snippet(editor_ref, iter, snippet):
    editor = editor_ref()
    buffer = editor.buffer
    view = editor.view

    expand_tabs = view.get_insert_spaces_instead_of_tabs()
    tab_width = view.get_tab_width()
    indent = unicode(get_whitespace(iter))

    buffer.begin_user_action()

    if not iter_at_whitespace(iter):
        start = iter.copy()
        start.backward_chars(len(snippet.snippet))
        while not start.equal(iter):
            txt = start.get_text(iter).decode('utf-8')
            if snippet.snippet.startswith(txt):
                buffer.delete(start, iter)
                break

            start.forward_char()

    offset = get_iter_at_cursor(buffer).get_offset()

    body, stop_offsets, insert_offsets = snippet.get_body_and_offsets(
        indent, expand_tabs, tab_width)

    buffer.insert_at_cursor(body)
    buffer.end_user_action()

    stop_managers[buffer] = StopManager(editor_ref, offset, stop_offsets, insert_offsets)

def show_proposals(view, iter, contexts):
    completion = view.get_completion()
    completion_context = completion.create_context(iter)
    completion.show([completion_providers[c] for c in contexts], completion_context)


class StopManager(object):
    def __init__(self, editor_ref, offset, stop_offsets, insert_offsets):
        buf = self.buffer = editor_ref().buffer
        self.editor_ref = editor_ref
        self.start_mark = buf.create_mark(None, buf.get_iter_at_offset(offset), True)
        self.end_mark = buf.create_mark(None, get_iter_at_cursor(buf))

        self.stop_marks = {}
        for i in sorted(stop_offsets):
            s, e = stop_offsets[i]
            s = buf.create_mark(None, buf.get_iter_at_offset(offset + s), True)
            e = buf.create_mark(None, buf.get_iter_at_offset(offset + e))
            self.stop_marks[i] = s, e

        self.insert_marks = {}
        for i in sorted(insert_offsets):
            s, e = insert_offsets[i]
            s = buf.create_mark(None, buf.get_iter_at_offset(offset + s), True)
            e = buf.create_mark(None, buf.get_iter_at_offset(offset + e))
            self.insert_marks[i] = s, e

        try:
            self.goto_stop(min(self.stop_marks))
        except ValueError:
            pass

    def goto_stop(self, idx):
        self.buffer.select_range(*reversed(self.get_iter_pair(*self.stop_marks[idx])))
        self.editor_ref().view.scroll_mark_onscreen(self.buffer.get_insert())

    def get_iter_pair(self, start_mark, end_mark):
        return (self.buffer.get_iter_at_mark(start_mark),
            self.buffer.get_iter_at_mark(end_mark))

    def cursor_in_snippet_range(self, cursor):
        return self.in_range(cursor, *self.get_iter_pair(self.start_mark, self.end_mark))

    def snippet_collapsed(self):
        s, e = self.get_iter_pair(self.start_mark, self.end_mark)
        return s.equal(e)

    def in_range(self, cursor, start, end):
        return cursor.in_range(start, end) or cursor.equal(end)

    def get_current_stop_idx(self, cursor):
        for i, (s, e) in self.stop_marks.iteritems():
            if self.in_range(cursor, *self.get_iter_pair(s, e)):
                return i

        return None

    def goto_next_stop(self, back=False):
        if self.buffer.get_has_selection():
            cursor = self.buffer.get_selection_bounds()[1]
        else:
            cursor = get_iter_at_cursor(self.buffer)

        idx = self.get_current_stop_idx(cursor)
        if idx is not None:
            try:
                if back:
                    idx = max(i for i in self.stop_marks if i < idx)
                else:
                    idx = min(i for i in self.stop_marks if i > idx)
            except ValueError:
                if back:
                    self.buffer.place_cursor(self.buffer.get_iter_at_mark(self.start_mark))
                else:
                    self.buffer.place_cursor(self.buffer.get_iter_at_mark(self.end_mark))

                self.editor_ref().view.scroll_mark_onscreen(self.buffer.get_insert())
                self.remove()
                return True

            self.goto_stop(idx)
            return True

        return False

    def replace_inserts(self):
        cursor = get_iter_at_cursor(self.buffer)
        if not cursor.equal(self.buffer.get_iter_at_mark(self.end_mark)):
            idx = self.get_current_stop_idx(cursor)
            if idx is not None:
                if idx in self.insert_marks:
                    txt = self.buffer.get_text(*self.get_iter_pair(*self.stop_marks[idx]))
                    s, e = self.insert_marks[idx]
                    self.buffer.handler_block_by_func(on_buffer_changed)
                    self.buffer.begin_user_action()
                    self.buffer.delete(*self.get_iter_pair(s, e))
                    self.buffer.insert(self.buffer.get_iter_at_mark(s), txt)
                    self.buffer.end_user_action()
                    self.buffer.handler_unblock_by_func(on_buffer_changed)

                return

        self.remove()

    def remove(self):
        self.editor_ref().message('Snippet was completed')

        try:
            del stop_managers[self.buffer]
        except KeyError:
            pass


class SnippetProposal(gobject.GObject, CompletionProposal):
    def __init__(self, snippet):
        gobject.GObject.__init__(self)
        self.snippet = snippet

    def do_get_label(self):
        return self.snippet.label

    def do_get_text(self):
        return self.snippet.label

    def do_get_info(self):
        return self.snippet.comment

def iter_at_whitespace(iter):
    if iter.starts_line():
        return True
    else:
        start = iter.copy()
        start.set_line(iter.get_line())

        char = start.get_text(iter).decode('utf-8')[-1]
        return not char.isalnum() and char != '_'


class SnippetsCompletionProvider(gobject.GObject, CompletionProvider):
    def __init__(self, ctx):
        gobject.GObject.__init__(self)
        self.ctx = ctx
        self.last_editor_ref = None

    def do_get_name(self):
        return '%s snippets' % self.ctx

    def do_get_priority(self):
        return 2

    def do_set_priority(self):
        pass

    def do_get_activation(self):
        return COMPLETION_ACTIVATION_USER_REQUESTED

    def do_populate(self, context):
        snippets = []
        all_snippets = sorted(loaded_snippets[self.ctx].values(), key=lambda r:r.label)
        iter = context.get_iter()

        if context.get_activation() == COMPLETION_ACTIVATION_USER_REQUESTED:
            if iter_at_whitespace(iter):
                snippets = all_snippets
            else:
                names = snippets_match_hash.get(self.ctx, {})
                if names:
                    already_added = {}
                    for cnt in range(max(names), 0, -1):
                        end = iter.copy()
                        end.backward_chars(cnt)

                        match = end.get_slice(iter)
                        for s in (s for s in all_snippets if s not in already_added):
                            if s.snippet.startswith(match):
                                already_added[s] = True
                                snippets.append(s)
        else:
            match = get_match(iter, self.ctx)
            snippets = [s for s in all_snippets if match == s.snippet]

        if snippets:
            context.add_proposals(self, [SnippetProposal(s) for s in snippets], True)
            self.last_editor_ref = context.props.completion.props.view.editor_ref
        else:
            context.add_proposals(self, [], True)

    def do_activate_proposal(self, proposal, iter):
        insert_snippet(self.last_editor_ref, iter, proposal.snippet)
        return True

gobject.type_register(SnippetsCompletionProvider)
gobject.type_register(SnippetProposal)



import os
from os.path import join, isdir, exists

from glib import markup_escape_text

from snaked.util import BuilderAware, join_to_file_dir, idle 
from snaked.core.shortcuts import ShortcutActivator

from snaked.core.plugins import get_plugin

def discover_plugins():
    import snaked.plugins
    result = {}
    for p in snaked.plugins.__path__:
        for n in os.listdir(p):         
            if isdir(join(p, n)) and exists(join(p, n, '__init__.py')):
                result[n] = True
                
    return result.keys()


class PluginDialog(BuilderAware):
    def __init__(self):
        BuilderAware.__init__(self, join_to_file_dir(__file__, 'plugin_prefs.glade'))
        self.activator = ShortcutActivator(self.window)
        self.activator.bind('Escape', self.hide)
        
        self.plugins_tree.get_columns()[0].get_cell_renderers()[0].props.activatable = True

    def show(self, enabled_plugins, callback):
        self.callback = callback
        self.fill_plugin_list(enabled_plugins)
        self.window.present()
        
    def hide(self):
        enabled_plugins = [p[1] for p in self.plugins if p[0]]
        self.callback(enabled_plugins)
        self.window.destroy()

    def on_delete_event(self, *args):
        idle(self.hide)
        return True        
        
    def get_aviable_plugins(self):
        for pname in discover_plugins():
            package = get_plugin(pname)
            name = getattr(package, 'name', pname)
            desc = getattr(package, 'desc', 'Some weird plugin')
            markup = "<b>%s</b>\n<small>%s</small>" % tuple(
                map(markup_escape_text, (name, desc)))
            
            yield (pname, markup, name)
    
    def fill_plugin_list(self, enabled):
        self.plugins.clear()
        
        for p in sorted(self.get_aviable_plugins(), key=lambda r: r[-1]):
            self.plugins.append((p[0] in enabled,) + p)

        self.plugins_tree.columns_autosize()
        self.plugins_tree.set_cursor("0", self.plugins_tree.get_columns()[0])
        
    def on_enabled_toggled(self, renderer, path):
        iter = self.plugins.get_iter(path)
        self.plugins.set_value(iter, 0, not renderer.get_active())
        


import gtk
from gobject import timeout_add, source_remove

class EscapeObject(object): pass

class FeedbackPopup(object):
    def __init__(self):

        self.window = gtk.Window(gtk.WINDOW_POPUP)
        self.window.set_property('allow-shrink', True)

        self.bar = gtk.EventBox()
        self.bar.set_border_width(5)

        box = gtk.EventBox()

        self.label = gtk.Label()
        self.label.set_alignment(0, 0)
        self.label.set_padding(10, 0)
        box.add(self.label)

        self.bar.add(box)
        self.bar.show_all()

        self.window.add(self.bar)

        self.timeout_id = None

        self.escape = None

    def remove_timeout(self):
        if self.timeout_id:
            source_remove(self.timeout_id)

        self.timeout_id = None

    def show(self, editor, text, timeout=1500):
        self.remove_timeout()
        if self.escape:
            self.hide()

        self.label.set_text(text)
        self.window.resize(*self.window.size_request())

        win = editor.view.get_window(gtk.TEXT_WINDOW_TEXT)
        x, y, w, h, _ = win.get_geometry()
        x, y = win.get_origin()
        mw, mh = self.window.get_size()

        self.window.move(x + w - mw, y + h - mh)
        self.window.show()

        if not self.escape:
            self.escape = EscapeObject()

        self.timeout_id = timeout_add(timeout, self.hide)

    def hide(self, *args):
        self.remove_timeout()
        self.window.hide()
        self.escape = None
        return False



import os.path
import time

import weakref

import gtk
import gtksourceview2

from ..util import save_file, idle, get_project_root, single_ref
from ..signals import SignalManager, Signal, connect_all, connect_external

class Editor(SignalManager):
    editor_closed = Signal()
    request_to_open_file = Signal(str, object, return_type=object)
    request_close = Signal()
    settings_changed = Signal()
    get_title = Signal(return_type=str)
    get_file_position = Signal(return_type=int)
    before_close = Signal()
    file_loaded = Signal()
    change_title = Signal(str)
    request_transient_for = Signal(object)
    file_saved = Signal()
    push_escape_callback = Signal(object, object)
    plugins_changed = Signal()
    add_spot_request = Signal()

    def __init__(self):
        self.uri = None
        self.session = None
        self.saveable = True
        self.lang = None
        self.contexts = []
        self.prefs = {}

        self.last_cursor_move = None

        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)

        self.buffer = gtksourceview2.Buffer()

        self.view = gtksourceview2.View()
        self.view.set_buffer(self.buffer)
        sw.add(self.view)
        self.view.editor_ref = weakref.ref(self)

        self.widget = gtk.VBox(False, 0)
        self.widget.pack_start(sw)

        self.widget.show_all()

        connect_all(self, buffer=self.buffer, view=self.view)

    def update_title(self):
        modified = '*' if self.buffer.get_modified() else ''

        if self.uri:
            title = self.get_title.emit()

            if not title:
                title = os.path.basename(self.uri)
        else:
            title = 'Unknown'

        self.change_title.emit(modified + title)

    def load_file(self, filename, line=None):
        self.uri = os.path.abspath(filename)
        self.encoding = 'utf-8'

        if os.path.exists(self.uri):
            self.view.window.freeze_updates()
            self.on_modified_changed_handler.block()
            self.buffer.begin_not_undoable_action()

            text = open(filename).read()

            try:
                utext = text.decode('utf-8')
            except UnicodeDecodeError, e:
                try:
                    import chardet
                    result = chardet.detect(text)
                    utext = text.decode(result['encoding'])
                    self.encoding = result['encoding']
                    idle(self.message, 'Automatically selected ' + self.encoding + 'encoding', 5000)
                except ImportError:
                    self.saveable = False
                    utext = str(e)
                except UnicodeDecodeError, ee:
                    self.saveable = False
                    utext = str(ee)

            self.buffer.set_text(utext)
            self.buffer.set_modified(False)

            self.buffer.end_not_undoable_action()
            self.on_modified_changed_handler.unblock()
            self.view.window.thaw_updates()

            pos = line if line is not None else self.get_file_position.emit()
            if pos is not None and pos >= 0:
                self.buffer.place_cursor(self.buffer.get_iter_at_line(pos))
                self.view.scroll_to_mark(self.buffer.get_insert(), 0.001, use_align=True, xalign=1.0)
            else:
                self.buffer.place_cursor(self.buffer.get_start_iter())

        self.file_loaded.emit()

        self.update_title()

    @connect_external('buffer', 'modified-changed', idle=True)
    def on_modified_changed(self, *args):
        self.update_title()

    def save(self):
        if not self.saveable:
            self.message("This file was opened with error and can't be saved")
            return

        if self.uri:
            if self.prefs['remove-trailing-space']:
                from snaked.core.processors import remove_trailing_spaces
                remove_trailing_spaces(self.buffer)

            try:
                save_file(self.uri, self.utext, self.encoding)
                if not self.buffer.get_modified():
                    self.message("%s saved" % self.uri)
                self.buffer.set_modified(False)
                self.file_saved.emit()
            except Exception, e:
                self.message(str(e), 5000)

    @property
    def project_root(self):
        if self.uri:
            root = get_project_root(self.uri)
            if not root:
                root = os.path.dirname(self.uri)

            return root

        return None

    def open_file(self, filename, line=None):
        return self.request_to_open_file.emit(filename, line)

    @property
    def cursor(self):
        """Return buffer's cursor iter

        :rtype: gtk.TextIter
        """
        return self.buffer.get_iter_at_mark(self.buffer.get_insert())

    @property
    def text(self):
        """Return buffer's content as utf-8 encoded string

        :rtype: str
        """
        return self.buffer.get_text(*self.buffer.get_bounds())

    @property
    def utext(self):
        """Return buffer's content as unicode string

        :rtype: unicode
        """
        return unicode(self.buffer.get_text(*self.buffer.get_bounds()), 'utf-8')

    def goto_line(self, line):
        iterator = self.buffer.get_iter_at_line(line - 1)
        self.buffer.place_cursor(iterator)
        self.view.scroll_to_mark(self.buffer.get_insert(), 0.001, use_align=True, xalign=1.0)

    def scroll_to_cursor(self):
        self.view.scroll_to_mark(self.buffer.get_insert(), 0.001, use_align=True, xalign=1.0)

    @single_ref
    def feedback_popup(self):
        from .feedback import FeedbackPopup
        return FeedbackPopup()

    def message(self, message, timeout=1500):
        popup = self.feedback_popup
        popup.show(self, message, timeout)
        self.push_escape(popup.hide, popup.escape)

    def push_escape(self, callback, *args):
        self.push_escape_callback.emit(callback, args)

    def add_spot(self):
        self.add_spot_request.emit()

    @connect_external('view', 'move-cursor')
    def on_cursor_moved(self, view, step_size, count, extend_selection):
        if not extend_selection:
            if step_size in (gtk.MOVEMENT_PAGES, gtk.MOVEMENT_BUFFER_ENDS):
                if self.last_cursor_move is None or time.time() - self.last_cursor_move > 7:
                    self.add_spot()

                self.last_cursor_move = time.time()
            elif step_size in (gtk.MOVEMENT_VISUAL_POSITIONS, ):
                self.last_cursor_move = None

    @connect_external('buffer', 'changed')
    def on_buffer_changed(self, buffer):
        self.last_cursor_move = None

    @connect_external('view', 'button-press-event')
    def on_button_press_event(self, view, event):
        if event.button == 1:
            return True

        return False


author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Python support'
desc = 'Autocompletion, definitions navigation and smart ident'

langs = ['python']

import weakref

handlers = weakref.WeakKeyDictionary()
outline_dialog = None

def init(manager):
    manager.add_shortcut('python-goto-definition', 'F3', 'Python',
        'Navigates to python definition', goto_definition)

    manager.add_shortcut('python-outline', '<ctrl>o', 'Python',
        'Opens outline dialog', open_outline)

    manager.add_shortcut('python-calltip', '<ctrl>Return', 'Python',
        'Shows calltips', show_calltips)

    from snaked.core.prefs import register_dialog
    register_dialog('Rope hints', edit_rope_hints, 'rope', 'hints')
    register_dialog('Rope config', edit_rope_config, 'rope', 'config')

def editor_created(editor):
    editor.connect('get-title', on_editor_get_title)

def editor_opened(editor):
    from plugin import Plugin
    h = Plugin(editor)
    handlers[editor] = h

def editor_closed(editor):
    del handlers[editor]

def quit():
    global outline_dialog
    if outline_dialog:
        outline_dialog.window.destroy()
        del outline_dialog

    import plugin
    for v in plugin.project_managers.values():
        v.close()

def goto_definition(editor):
    try:
        h = handlers[editor]
    except KeyError:
        return

    h.goto_definition()

def show_calltips(editor):
    try:
        h = handlers[editor]
    except KeyError:
        return

    h.show_calltips()

def on_editor_get_title(editor):
    if editor.uri.endswith('.py'):
        return get_python_title(editor.uri)

    return None

def get_python_title(uri):
    from os.path import dirname, basename, exists, join

    title = basename(uri)
    packages = []
    while True:
        path = dirname(uri)
        if path == uri:
            break

        uri = path

        if exists(join(uri, '__init__.py')):
            packages.append(basename(uri))
        else:
            break

    if packages:
        if title != '__init__.py':
            packages.insert(0, title.partition('.py')[0])

        return '.'.join(reversed(packages))
    else:
        return None

def open_outline(editor):
    global outline_dialog
    if not outline_dialog:
        from outline import OutlineDialog
        outline_dialog = OutlineDialog()

    outline_dialog.show(editor)

def edit_rope_hints(editor):
    import os
    import shutil
    from os.path import join, exists, dirname

    ropehints = join(editor.project_root, '.ropeproject', 'ropehints.py')
    if not exists(ropehints):
        ropedir = dirname(ropehints)
        if not exists(ropedir):
            os.mkdir(ropedir, 0755)

        shutil.copy(join(dirname(__file__), 'ropehints_tpl.py'), ropehints)

    editor.open_file(ropehints)

def edit_rope_config(editor):
    from os.path import join, exists
    ropeconfig = join(editor.project_root, '.ropeproject', 'config.py')
    if exists(ropeconfig):
        editor.open_file(ropeconfig)
    else:
        editor.message('There is no existing rope config.\n'
            'Try to autocomplete something. Then repeat', 5000)



import weakref

from gtk import accelerator_name, accel_map_change_entry
from pango import WEIGHT_BOLD, WEIGHT_NORMAL

from snaked.util import BuilderAware, join_to_file_dir, idle
from snaked.core import shortcuts

class ShortcutsDialog(BuilderAware):
    def __init__(self):
        BuilderAware.__init__(self, join_to_file_dir(__file__, 'shortcuts.glade'))
        self.activator = shortcuts.ShortcutActivator(self.window)
        self.activator.bind('Escape', self.hide)
        
    def show(self, editor):
        self.editor = weakref.ref(editor)        
        self.fill_actions()
        editor.request_transient_for.emit(self.window)
        self.window.present()
    
    def get_weight(self, path, key, mod):
        return WEIGHT_BOLD if shortcuts.default_shortcuts[path] != (key, mod) else WEIGHT_NORMAL
        
    def fill_actions(self):
        self.actions_view.set_model(None)

        categories = {}
        sortkey = lambda v: v[0]
        
        for path, key, mod in sorted(shortcuts.get_registered_shortcuts(), key=sortkey):
            _, category, name = path.split('/')
            
            if category not in categories:
                categories[category] = self.actions.append(
                    None, (category, None, None, False, WEIGHT_NORMAL))
            
            accel = accelerator_name(key, mod)
            desc = shortcuts.registered_shortcuts[name][1]
            
            self.actions.append(categories[category], (name, accel, desc, True,
                self.get_weight(path, key, mod)))    
        
        self.actions_view.set_model(self.actions)
        self.actions_view.expand_all()
        
    def hide(self):
        shortcuts.save_shortcuts()
        self.editor().message('Key configuration saved')
        self.window.destroy()

    def on_delete_event(self, *args):
        idle(self.hide)
        return True
        
    def on_accel_edited(self, renderer, path, accel_key, accel_mods, hardware_keycode):
        iter = self.actions.get_iter(path)
        accel = accelerator_name(accel_key, accel_mods)
        name = self.actions.get_value(iter, 0)
        path = shortcuts.get_path_by_name(name)

        accel_map_change_entry(path, accel_key, accel_mods, True)
        shortcuts.names_by_key.clear()
        
        self.actions.set_value(iter, 1, accel)
        self.actions.set_value(iter, 4, self.get_weight(path, accel_key, accel_mods))
    
    def on_accel_cleared(self, renderer, path, *args):
        iter = self.actions.get_iter(path)
        name = self.actions.get_value(iter, 0)
        path = shortcuts.get_path_by_name(name)

        accel_map_change_entry(path, 0, 0, True)
        shortcuts.names_by_key.clear()
        
        self.actions.set_value(iter, 1, None)
        self.actions.set_value(iter, 4, WEIGHT_BOLD)



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Goto line'
desc = 'Navigates to specified line'

import gtk

from snaked.util import idle

def init(manager):
    manager.add_shortcut('goto-line', '<ctrl>l', 'Edit', 'Goto line', goto_line)

def goto_line(editor):
    widget = get_widget(editor)
    
    editor.widget.pack_start(widget, False)
    widget.entry.grab_focus()
    widget.show_all()

def get_widget(editor):
    widget = gtk.HBox(False, 0)
    
    label = gtk.Label()
    label.set_text('Goto line:')
    widget.pack_start(label, False)
    
    entry = gtk.Entry()
    widget.pack_start(entry, False)
    entry.connect('activate', on_entry_activate, editor, widget)
    entry.connect('focus-out-event', on_focus_out, editor, widget)
    entry.connect('key-press-event', on_key_press, editor, widget)
    
    widget.entry = entry
    
    return widget
    
def hide(editor, widget):
    if widget and widget.get_parent():
        editor.widget.remove(widget)
        widget.destroy()
    editor.view.grab_focus()

def on_focus_out(sender, event, editor, widget):
    idle(hide, editor, widget)
    
def on_entry_activate(sender, editor, widget):
    idle(hide, editor, widget)
    try:
        line = int(sender.get_text())
        editor.add_spot()
        idle(editor.goto_line, line)
    except ValueError:
        pass

def on_key_press(sender, event, editor, widget):
    if event.keyval == gtk.keysyms.Escape:
        idle(hide, editor, widget)
        return True
    
    return False


import textwrap

import weakref

import gtk
from gtksourceview2 import CompletionProvider, CompletionProposal
from gtksourceview2 import COMPLETION_ACTIVATION_USER_REQUESTED

import gobject
from glib import markup_escape_text

def pangonify_rst(text):
    result = ''

    lines = text.strip().expandtabs().splitlines()
    indent = 1000
    for l in lines[1:]:
        stripped = l.lstrip()
        if stripped:
            indent = min(indent, len(l) - len(stripped))

    trimmed = [lines[0].strip()]
    if indent < 1000:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())

    wrapper = textwrap.TextWrapper(width=60)
    paragraph = ''
    break_paragraph = False

    def add_paragraph():
        r = result
        if paragraph:
            if break_paragraph:
                r += '\n'

            r += wrapper.fill(paragraph) + '\n\n'
            return r, '', False

        return r, '', break_paragraph

    for l in trimmed:
        if not l.strip():
            result, paragraph, break_paragraph = add_paragraph()
        else:
            if l.startswith('  '):
                result, paragraph, break_paragraph = add_paragraph()
                break_paragraph = True
                result += "<tt>%s</tt>" % markup_escape_text(l)
                result += '\n'
            else:
                paragraph += markup_escape_text(l) + ' '

    if paragraph:
        result, _, _ = add_paragraph()

    return result


class Proposal(gobject.GObject, CompletionProposal):
    def __init__(self, proposal):
        gobject.GObject.__init__(self)
        self.proposal = proposal

    def do_get_label(self):
        return self.proposal.name

    def do_get_text(self):
        return self.proposal.name

    def do_get_info(self):
        info = self.proposal.get_doc()
        if info:
            return pangonify_rst(info)
        else:
            return ''

class RopeCompletionProvider(gobject.GObject, CompletionProvider):
    def __init__(self, plugin):
        gobject.GObject.__init__(self)
        self.plugin = weakref.ref(plugin)
        self.info_widget = gtk.ScrolledWindow()
        self.info_widget.set_size_request(400, 300)
        self.info_widget.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        label = gtk.Label()
        label.set_alignment(0, 0)
        label.set_line_wrap(False)
        self.info_widget.add_with_viewport(label)
        self.info_widget.label = label
        self.info_widget.show_all()

    def do_get_name(self):
        return 'python'

    def do_get_priority(self):
        return 1

    def do_set_priority(self):
        pass

    def do_get_activation(self):
        return COMPLETION_ACTIVATION_USER_REQUESTED

    def do_get_info_widget(self, proposal):
        self.info_widget.label.set_markup(proposal.get_info())
        return self.info_widget

    def do_update_info(self, proposal, info):
        info.get_widget().label.set_markup(proposal.get_info())

    def do_populate(self, context):
        project = self.plugin().project_manager.project

        from rope.contrib import codeassist

        try:
            proposals = codeassist.sorted_proposals(
                codeassist.code_assist(project, *self.plugin().get_source_and_offset(),
                    resource=self.plugin().get_rope_resource(project), maxfixes=3))

        except Exception, e:
            import traceback
            traceback.print_exc()
            self.plugin().editor.message(str(e), 5000)
            return

        if proposals:
            context.add_proposals(self, [Proposal(p) for p in proposals], True)
        else:
            context.add_proposals(self, [], True)
            self.plugin().editor.message("Can't assist")

gobject.type_register(RopeCompletionProvider)
gobject.type_register(Proposal)



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Complete words'
desc = 'Cycle through possible word completions'

import weakref

from gobject import timeout_add, source_remove
from string import whitespace

from snaked.util import idle, refresh_gui
from snaked.signals import connect_all, connect_external

editors_to_update = []
timer_source_id = None
handlers = weakref.WeakKeyDictionary()

def init(manager):
    manager.add_shortcut('complete-word', '<alt>slash', 'Edit',
        'Cycle through word completions', complete_word)
    
    global timer_source_id
    timer_source_id = timeout_add(3000, update_words_timer)

def complete_word(editor):
    try:
        h = handlers[editor]
    except KeyError:
        return
        
    h.cycle()

def editor_opened(editor):
    idle(add_update_job, editor)
    h = Plugin(editor)
    handlers[editor] = h

def editor_closed(editor):
    try:
        del handlers[editor]
    except KeyError:
        pass
            
def quit():
    source_remove(timer_source_id)

def add_update_job(editor):
    import words
    words.add_job(editor.uri, editor.text)

def update_words_timer():
    if editors_to_update:
        for e in editors_to_update:
            add_update_job(e)
        
        editors_to_update[:] = []
    
    return True


class Plugin(object):
    def __init__(self, editor):
        self.editor = editor
        connect_all(self, buffer=editor.buffer)
        self.start_word = None
        self.start_offset = None
    
    @connect_external('buffer', 'changed', idle=True)    
    def on_buffer_changed(self, *args):
        if not self.editor in editors_to_update:
            editors_to_update.append(self.editor)
            
        self.start_word = None
        self.start_iter = None

    def is_valid_character(self, c):
        if c in whitespace:
            return False
        
        return c.isalpha() or c.isdigit() or (c in ("-", "_"))

    def backward_to_word_begin(self, iterator):
        if iterator.starts_line(): return iterator
        iterator.backward_char()
        while self.is_valid_character(iterator.get_char()):
            iterator.backward_char()
            if iterator.starts_line(): return iterator
        iterator.forward_char()
        return iterator

    def forward_to_word_end(self, iterator):
        if iterator.ends_line(): return iterator
        if not self.is_valid_character(iterator.get_char()): return iterator
        while self.is_valid_character(iterator.get_char()):
            iterator.forward_char()
            if iterator.ends_line(): return iterator
        return iterator
    
    def get_word_before_cursor(self):
        iterator = self.editor.cursor
        # If the cursor is in front of a valid character we ignore
        # word completion.
        if self.is_valid_character(iterator.get_char()):
            return None, None
        
        if iterator.starts_line():
            return None, None
    
        iterator.backward_char()
        
        if not self.is_valid_character(iterator.get_char()):
            return None, None
    
        start = self.backward_to_word_begin(iterator.copy())
        end = self.forward_to_word_end(iterator.copy())
        word = self.editor.buffer.get_text(start, end).strip()
        
        return word, start

    def get_matches(self, string):
        import words
        
        if not words.words:
            return None
        
        result = []
        for word, files in words.words.iteritems():
            if word != string and word.startswith(string):
                result.append((word, sum(files.values())))
                
        result.sort(key=lambda r: r[1], reverse=True)
        
        return [r[0] for r in result]

    def cycle(self):
        word_to_complete, start = self.get_word_before_cursor()
        
        if not word_to_complete:
            return False

        if not self.start_word or self.start_offset != start.get_offset():
            self.start_word = word_to_complete
            self.start_offset = start.get_offset()

        matches = self.get_matches(self.start_word)
        if matches:
            idx = 0
            try:
                idx = matches.index(word_to_complete)
                idx = (idx + 1) % len(matches)
            except ValueError:
                pass

            if matches[idx] == word_to_complete:
                self.editor.message("Word completed already")
                return False
                
            self.on_buffer_changed_handler.block()             
            
            end = self.editor.cursor
            self.editor.buffer.delete(start, end)
            self.editor.buffer.insert(start, matches[idx])
            
            refresh_gui()
            self.on_buffer_changed_handler.unblock()             
        else:
            self.editor.message("No word to complete")



import os.path
import weakref

import gtk
import gtksourceview2
import pango

from ..signals import connect_all
from ..util import idle, lazy_property, create_lang_matchers_from_file, LangGuesser, get_project_root

import prefs
from .shortcuts import register_shortcut, load_shortcuts
from .plugins import PluginManager
from .editor import Editor

import snaked.core.quick_open


class EditorManager(object):
    def __init__(self):
        self.editors = []
        self.style_manager = gtksourceview2.style_scheme_manager_get_default()
        self.lang_manager = gtksourceview2.language_manager_get_default()

        self.plugin_manager = PluginManager()
        prefs.register_dialog('Plugins', self.plugin_manager.show_plugins_prefs, 'plugin',
            'extension')

        prefs.register_dialog('Key configuration', self.show_key_preferences, 'key', 'bind', 'shortcut')

        prefs.register_dialog('Editor settings', self.show_editor_preferences,
            'editor', 'font', 'style', 'margin', 'line', 'tab', 'whitespace')

        self.escape_stack = []
        self.escape_map = {}
        self.spot_history = []
        self.lang_gussers = {}

        self.session = None

        load_shortcuts()
        self.register_app_shortcuts()

        # Init core plugins
        self.plugin_manager.load_core_plugin(snaked.core.quick_open)


    def register_app_shortcuts(self):
        register_shortcut('quit', '<ctrl>q', 'Application', 'Quit')
        register_shortcut('close-window', '<ctrl>w', 'Window', 'Closes window')
        register_shortcut('save', '<ctrl>s', 'File', 'Saves file')
        register_shortcut('new-file', '<ctrl>n', 'File',
            'Open dialog to choose new file directory and name')
        register_shortcut('show-preferences', '<ctrl>p', 'Window', 'Open preferences dialog')

        register_shortcut('place-spot', '<alt>t', 'Edit', 'Place spot at current cursor location')
        register_shortcut('goto-last-spot', '<alt>q', 'Edit', 'Quick jump to last placed spot')
        register_shortcut('goto-next-spot', '<ctrl><alt>Right', 'Edit',
            'Quick jump to next spot in history')
        register_shortcut('goto-prev-spot', '<ctrl><alt>Left', 'Edit',
            'Quick jump to previous spot in history')

    def get_lang_guesser(self, project_root):
        try:
            return self.lang_gussers[project_root]
        except KeyError:
            pass

        guesser = None
        if project_root:
            contexts_filename = os.path.join(project_root, '.snaked_project', 'contexts')
            if os.path.exists(contexts_filename):
                matchers = create_lang_matchers_from_file(project_root, contexts_filename)
                guesser = LangGuesser(matchers)

        self.lang_gussers[project_root] = guesser
        return guesser

    def open(self, filename, line=None):
        editor = Editor()
        self.editors.append(editor)
        editor.session = self.session

        connect_all(self, editor)

        idle(self.set_editor_prefs, editor, filename)
        idle(self.set_editor_shortcuts, editor)
        idle(self.plugin_manager.editor_created, editor)

        self.manage_editor(editor)

        idle(editor.load_file, filename, line)
        idle(self.plugin_manager.editor_opened, editor)

        return editor

    @lazy_property
    def lang_prefs(self):
        return prefs.load_json_settings('langs.conf', {})

    def set_editor_prefs(self, editor, filename):
        lang = None
        editor.lang = 'default'
        editor.contexts = ['default']

        if filename:
            root = get_project_root(filename)
            guesser = self.get_lang_guesser(root)
            if guesser:
                contexts = guesser.guess(os.path.abspath(filename))
                if contexts:
                    lang_id = contexts[0]
                    lang = self.lang_manager.get_language(lang_id)
                    if lang:
                        editor.lang = lang.get_id()
                        editor.contexts = contexts

            if not lang:
                lang = self.lang_manager.guess_language(filename, None)
                if lang:
                    editor.lang = lang.get_id()
                    editor.contexts = [editor.lang]

        if lang:
            editor.buffer.set_language(lang)

        pref = prefs.CompositePreferences(self.lang_prefs.get(editor.lang, {}),
            self.lang_prefs.get('default', {}), prefs.default_prefs.get(editor.lang, {}),
            prefs.default_prefs['default'])

        style_scheme = self.style_manager.get_scheme(pref['style'])
        editor.buffer.set_style_scheme(style_scheme)

        font = pango.FontDescription(pref['font'])
        editor.view.modify_font(font)

        editor.view.set_auto_indent(pref['auto-indent'])
        editor.view.set_indent_on_tab(pref['indent-on-tab'])
        editor.view.set_insert_spaces_instead_of_tabs(not pref['use-tabs'])
        editor.view.set_smart_home_end(pref['smart-home-end'])
        editor.view.set_highlight_current_line(pref['highlight-current-line'])
        editor.view.set_show_line_numbers(pref['show-line-numbers'])
        editor.view.set_tab_width(pref['tab-width'])
        editor.view.set_draw_spaces(pref['show-whitespace'])
        editor.view.set_right_margin_position(pref['right-margin'])
        editor.view.set_show_right_margin(pref['show-right-margin'])
        editor.view.set_wrap_mode(gtk.WRAP_WORD if pref['wrap-text'] else gtk.WRAP_NONE)
        editor.view.set_pixels_above_lines(pref['line-spacing'])

        editor.prefs = pref

    @Editor.editor_closed(idle=True)
    def on_editor_closed(self, editor):
        self.plugin_manager.editor_closed(editor)
        self.editors.remove(editor)

        if not self.editors:
            snaked.core.quick_open.activate(self.get_fake_editor())

    @Editor.change_title
    def on_editor_change_title(self, editor, title):
        self.set_editor_title(editor, title)

    @Editor.request_close
    def on_editor_close_request(self, editor):
        self.close_editor(editor)

    @Editor.request_to_open_file
    def on_request_to_open_file(self, editor, filename, line):
        self.add_spot(editor)

        for e in self.editors:
            if e.uri == filename:
                self.focus_editor(e)

                if line is not None:
                    e.goto_line(line + 1)

                break
        else:
            e = self.open(filename, line)

        return e

    @Editor.request_transient_for
    def on_request_transient_for(self, editor, window):
        self.set_transient_for(editor, window)

    @Editor.settings_changed(idle=True)
    def on_editor_settings_changed(self, editor):
        self.set_editor_prefs(editor, editor.uri)
        for e in self.editors:
            if e is not editor:
                idle(self.set_editor_prefs, e, e.uri)

    def new_file_action(self, editor):
        from snaked.core.gui import new_file
        new_file.show_create_file(editor)

    def quit(self, editor):
        if self.session:
            self.save_session(self.session, editor)

        map(self.plugin_manager.editor_closed, self.editors)

        self.plugin_manager.quit()

        if gtk.main_level() > 0:
            gtk.main_quit()

    def get_session_settings(self, session):
        from .prefs import load_json_settings
        return load_json_settings('%s.session' % session, {})

    def save_session(self, session, active_editor=None):
        from .prefs import save_json_settings
        settings = self.get_session_settings(session)
        settings['files'] = [e.uri for e in self.editors if e.uri]
        settings['active_file'] = active_editor.uri if active_editor else None
        save_json_settings('%s.session' % session, settings)

    @Editor.push_escape_callback
    def on_push_escape_callback(self, editor, callback, args):
        key = (callback,) + tuple(map(id, args))
        if key in self.escape_map:
            return

        self.escape_map[key] = True
        self.escape_stack.append((key, callback, map(weakref.ref, args)))

    def process_escape(self, editor):
        while self.escape_stack:
            key, cb, args = self.escape_stack.pop()
            del self.escape_map[key]
            realargs = [a() for a in args]
            if not any(a is None for a in realargs):
                cb(*realargs)
                return False

        return False

    def show_key_preferences(self, editor):
        from snaked.core.gui.shortcuts import ShortcutsDialog
        dialog = ShortcutsDialog()
        dialog.show(editor)

    def show_preferences(self, editor):
        from snaked.core.gui.prefs import PreferencesDialog
        dialog = PreferencesDialog()
        dialog.show(editor)

    def show_editor_preferences(self, editor):
        from snaked.core.gui.editor_prefs import PreferencesDialog
        dialog = PreferencesDialog(self.lang_prefs)
        dialog.show(editor)

    @Editor.plugins_changed
    def on_plugins_changed(self, editor):
        for e in self.editors:
            self.set_editor_shortcuts(e)

    def get_fake_editor(self):
        self.fake_editor = FakeEditor(self)
        return self.fake_editor

    def add_spot_with_feedback(self, editor):
        self.add_spot(editor)
        editor.message('Spot added')

    @Editor.add_spot_request
    def add_spot(self, editor):
        self.add_spot_to_history(EditorSpot(self, editor))

    def add_spot_to_history(self, spot):
        self.spot_history = [s for s in self.spot_history
            if s.is_valid() and not s.similar_to(spot)]

        self.spot_history.insert(0, spot)

        while len(self.spot_history) > 30:
            self.spot_history.pop()

    def goto_last_spot(self, back_to=None):
        new_spot = EditorSpot(self, back_to) if back_to else None
        spot = self.get_last_spot(new_spot)
        if spot:
            spot.goto(back_to)
            if new_spot:
                self.add_spot_to_history(new_spot)
        else:
            if back_to:
                back_to.message('Spot history is empty')

    def get_last_spot(self, exclude_spot=None, exclude_editor=None):
        for s in self.spot_history:
            if s.is_valid() and not s.similar_to(exclude_spot) and s.editor() is not exclude_editor:
                return s

        return None

    def goto_next_prev_spot(self, editor, is_next):
        current_spot = EditorSpot(self, editor)
        if is_next:
            seq = self.spot_history
        else:
            seq = reversed(self.spot_history)

        prev_spot = None
        for s in (s for s in seq if s.is_valid()):
            if s.similar_to(current_spot):
                if prev_spot:
                    prev_spot.goto(editor)
                else:
                    editor.message('No more spots to go')
                return

            prev_spot = s

        self.goto_last_spot(editor)

class EditorSpot(object):
    def __init__(self, manager, editor):
        self.manager = manager
        self.editor = weakref.ref(editor)
        self.mark = editor.buffer.create_mark(None, editor.cursor)

    @property
    def iter(self):
        return self.mark.get_buffer().get_iter_at_mark(self.mark)

    def is_valid(self):
        return self.editor() and not self.mark.get_deleted()

    def similar_to(self, spot):
        return spot and self.mark.get_buffer() is spot.mark.get_buffer() \
            and abs(self.iter.get_line() - spot.iter.get_line()) < 7

    def __del__(self):
        buffer = self.mark.get_buffer()
        if buffer:
            buffer.delete_mark(self.mark)

    def goto(self, back_to=None):
        editor = self.editor()
        editor.buffer.place_cursor(self.iter)
        editor.scroll_to_cursor()

        if editor is not back_to:
            self.manager.focus_editor(editor)


class FakeEditor(object):
    def __init__(self, manager):
        self.project_root = None
        self.manager = manager
        self.request_transient_for = self
        self.session = manager.session

    def emit(self, window):
        self.manager.set_transient_for(self, window)

    def open_file(self, filename, line=None):
        result = self.manager.open(filename, line)
        del self.manager.fake_editor
        self.manager.fake_editor = None
        return result

    def on_dialog_escape(self, dialog):
        self.manager.quit(None)

    def message(self, message, timeout=None):
        print message



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Search'
desc = 'Searches words in document'

import re
import weakref
import gtk
from snaked.util import idle

active_widgets = weakref.WeakKeyDictionary()
search_selections = []

class SearchSelection(object):
    def __init__(self, search):
        self.search = search

def init(manager):
    manager.add_shortcut('search', '<ctrl>f', 'Edit', 'Search', search)
    manager.add_shortcut('find-next', '<ctrl>j', 'Edit', 'Find next', find_next)
    manager.add_shortcut('find-prev', '<ctrl>k', 'Edit', 'Find prev', find_prev)
    manager.add_shortcut('mark-selection', '<ctrl>h', 'Edit',
        'Mark selection occurrences', mark_selection)

def search(editor):
    if editor in active_widgets:
        widget = active_widgets[editor]
        widget.entry.grab_focus()
    else:
        widget = create_widget(editor)
        active_widgets[editor] = widget
        editor.widget.pack_start(widget, False)
        widget.entry.grab_focus()
        editor.push_escape(hide, editor, widget)
        widget.show_all()

def backward_search(matcher, text, endpos):
    match = None
    for m in matcher.finditer(text):
        if m.end() > endpos:
            return match
        match = m

    return match

def do_find(editor, dir, start_from=None):
    if editor in active_widgets:
        search = active_widgets[editor].entry.get_text()
        ignore_case = active_widgets[editor].ignore_case.get_active()
        regex = active_widgets[editor].regex.get_active()
    elif search_selections:
        search = search_selections[0].search
        ignore_case = False
        regex = False
    else:
        return

    matcher = get_matcher(editor, search, ignore_case, regex)
    if not matcher:
        return

    iter = start_from
    if not iter:
        if editor.buffer.get_has_selection() and dir == 1:
            iter = editor.buffer.get_iter_at_mark(editor.buffer.get_selection_bound())
        else:
            iter = editor.cursor

    match = None
    if dir == 0:
        match = matcher.search(editor.utext, iter.get_offset())
    else:
        match = backward_search(matcher, editor.utext, iter.get_offset())

    if match:
        bounds = map(editor.buffer.get_iter_at_offset, match.span())
        editor.buffer.select_range(bounds[1], bounds[0])
        editor.view.scroll_to_iter(bounds[0], 0.001, use_align=True, xalign=1.0)
        if start_from:
            editor.message('Wrap search', 800)

        return True
    elif not start_from:
        return do_find(editor, dir, editor.buffer.get_bounds()[dir])
    else:
        editor.message('Text not found')

    return False

def find_prev(editor):
    do_find(editor, 1)

def find_next(editor, grab_focus=False):
    if do_find(editor, 0) and grab_focus:
        editor.view.grab_focus()

def create_widget(editor):
    widget = gtk.HBox(False, 10)
    widget.set_border_width(3)

    label = gtk.Label()
    label.set_text('_Search:')
    label.set_use_underline(True)
    widget.pack_start(label, False)

    entry = gtk.Entry()
    widget.pack_start(entry, False)
    widget.entry = entry
    label.set_mnemonic_widget(entry)
    entry.connect('activate', on_search_activate, editor, widget)


    label = gtk.Label()
    label.set_text('_Replace:')
    label.set_use_underline(True)
    widget.pack_start(label, False)

    entry = gtk.Entry()
    widget.pack_start(entry, False)
    widget.replace_entry = entry
    label.set_mnemonic_widget(entry)
    entry.connect('activate', on_search_activate, editor, widget)


    cb = gtk.CheckButton('Ignore _case')
    widget.pack_start(cb, False)
    widget.ignore_case = cb

    cb = gtk.CheckButton('Rege_x')
    widget.pack_start(cb, False)
    widget.regex = cb

    label = gtk.Label('Re_place')
    label.set_use_underline(True)
    label.set_padding(10, 1)
    button = gtk.Button()
    button.add(label)
    widget.pack_start(button, False)
    button.connect('activate', on_replace_activate, editor, widget)

    label = gtk.Label('Replace _all')
    label.set_use_underline(True)
    label.set_padding(10, 1)
    button = gtk.Button()
    button.add(label)
    widget.pack_start(button, False)
    button.connect('activate', on_replace_all_activate, editor, widget)

    return widget

def get_tag(editor):
    table = editor.buffer.get_tag_table()
    tag = table.lookup('search')
    if not tag:
        style = editor.buffer.get_style_scheme().get_style('search-match')
        tag = editor.buffer.create_tag('search')

        if style.props.background_set:
            tag.props.background = style.props.background

        if style.props.foreground_set:
            tag.props.foreground = style.props.foreground

    return tag

def delete_all_marks(editor):
    start, end = editor.buffer.get_bounds()
    if editor.buffer.get_tag_table().lookup('search'):
        editor.buffer.remove_tag_by_name('search', start, end)

def get_matcher(editor, search, ignore_case, regex):
    flags = re.UNICODE
    if ignore_case:
        flags |= re.IGNORECASE

    if regex:
        try:
            return re.compile(unicode(search), flags)
        except Exception, e:
            editor.message('Bad regex: ' + str(e), 3000)
            if editor in active_widgets:
                idle(active_widgets[editor].entry.grab_focus)
            return None
    else:
        return re.compile(re.escape(unicode(search)), flags)

def mark_occurences(editor, search, ignore_case, regex):
    matcher = get_matcher(editor, search, ignore_case, regex)
    if not matcher:
        return False

    count = 0
    for m in matcher.finditer(editor.utext):
        editor.buffer.apply_tag(get_tag(editor),
            *map(editor.buffer.get_iter_at_offset, m.span()))

        count += 1

    if count == 1:
        idle(editor.message, 'One occurrence is marked')
    elif count > 1:
        idle(editor.message, '%d occurrences are marked' % count)
    else:
        idle(editor.message, 'Text not found')
        return False

    return True

def on_search_activate(sender, editor, widget):
    delete_all_marks(editor)
    if mark_occurences(editor, widget.entry.get_text(),
            widget.ignore_case.get_active(), widget.regex.get_active()):
        find_next(editor, True)

def hide(editor, widget):
    delete_all_marks(editor)

    try:
        del active_widgets[editor]
    except KeyError:
        pass

    if widget and widget.get_parent():
        editor.widget.remove(widget)
        widget.destroy()

    editor.view.grab_focus()

def mark_selection(editor):
    if not editor.buffer.get_has_selection():
        editor.message('Select something')
        return

    if search_selections:
        search_selections[:] = []

    delete_all_marks(editor)

    occur = SearchSelection(editor.buffer.get_text(*editor.buffer.get_selection_bounds()))
    search_selections.append(occur)

    def remove_all(editor, occur):
        search_selections[:] = []
        delete_all_marks(editor)

    mark_occurences(editor, occur.search, False, False)
    editor.push_escape(remove_all, editor, occur)

def on_replace_activate(button, editor, widget):
    if editor not in active_widgets:
        return

    search = active_widgets[editor].entry.get_text()
    ignore_case = active_widgets[editor].ignore_case.get_active()
    regex = active_widgets[editor].regex.get_active()
    replace = unicode(active_widgets[editor].replace_entry.get_text())

    matcher = get_matcher(editor, search, ignore_case, regex)
    if not matcher:
        return

    if editor.buffer.get_has_selection():
        iter = editor.buffer.get_selection_bounds()[0]
    else:
        iter = editor.cursor

    match = matcher.search(editor.utext, iter.get_offset())
    if not match:
        editor.message('Replace what?')
        return

    start, end = map(editor.buffer.get_iter_at_offset, match.span())
    editor.buffer.begin_user_action()
    editor.buffer.place_cursor(start)
    editor.buffer.delete(start, end)
    editor.buffer.insert_at_cursor(match.expand(replace).encode('utf-8'))
    editor.buffer.end_user_action()

    idle(editor.scroll_to_cursor)

def on_replace_all_activate(button, editor, widget):
    if editor not in active_widgets:
        return

    search = active_widgets[editor].entry.get_text()
    ignore_case = active_widgets[editor].ignore_case.get_active()
    regex = active_widgets[editor].regex.get_active()
    replace = unicode(active_widgets[editor].replace_entry.get_text())

    matcher = get_matcher(editor, search, ignore_case, regex)
    if not matcher:
        return

    line, offset = editor.cursor.get_line(), editor.cursor.get_line_offset()

    editor.buffer.begin_user_action()
    editor.buffer.place_cursor(editor.buffer.get_bounds()[0])
    count = 0
    while True:
        match = matcher.search(editor.utext, editor.cursor.get_offset())
        if not match:
            break

        start, end = map(editor.buffer.get_iter_at_offset, match.span())
        editor.buffer.place_cursor(start)
        editor.buffer.delete(start, end)
        editor.buffer.insert_at_cursor(match.expand(replace).encode('utf-8'))
        count += 1

    editor.buffer.end_user_action()

    if not count:
        editor.message('Nothing to replace')
    elif count == 1:
        editor.message('One occurrence was replaced')
    else:
        editor.message('%d occurrences were replaced' % count)

    cursor = editor.cursor
    cursor.set_line(line)
    cursor.set_line_offset(offset)
    editor.buffer.place_cursor(cursor)
    editor.scroll_to_cursor()


import re

match_start = re.compile(r'"""|\'\'\'|"|\'|\(|\[|\{|\)|\]|\}')
match_quote_end = {
    '"""': re.compile(r'.*?"""'),
    "'''": re.compile(r".*?'''"),
    "'": re.compile(r"'|.*?[^\\]'"),
    '"': re.compile(r'"|.*?[^\\]"'),
}

inverted = {
    ')': '(',
    '}': '{',
    ']': '[',
}

def find_closing_quote_pos(quote, text, start):
    m = match_quote_end[quote].match(text, start)
    if m:
        return m.end()
        
    return None

def get_brackets(text, cursor):
    start = 0
    open = []
    close = []
    level = None
    while True:
        m = match_start.search(text, start)
        if not m:
            break
            
        start = m.end()
        bracket = m.group()

        if start > cursor:
            if not open:
                return None, None, None
            elif level is None:
                level = len(open) - 1

        if bracket in ['"""', "'''", '"', "'"]:
            end = find_closing_quote_pos(bracket, text, start)
            if not end:
                return None, None, None
            if level is None and end > cursor:
                return bracket, start, end
            else:
                start = end
        elif bracket in ['(', '[', '{']:
            open.append((bracket, start))
        elif open and bracket in [')', ']', '}']:
            close.append((inverted[bracket], start))
        
        while open and close and open[-1][0] == close[-1][0]:
            br, opos = open.pop()
            br, cpos = close.pop()
            
            if len(open) == level:
                return br, opos, cpos



import os

import weakref
import gtk
import gio

from snaked.signals import connect_external, connect_all, weak_connect
from snaked.util import idle, lazy_property

from .ropehints import CompositeHintProvider

project_managers = weakref.WeakValueDictionary()

class RopeProjectManager(object):
    def __init__(self, project):
        self.project = project
        self.hints_monitor = None

        if project.ropefolder:
            self.hints_filename = os.path.join(project.ropefolder.real_path, 'ropehints.py')
            self.hints_monitor = gio.File(self.hints_filename).monitor_file()
            weak_connect(self.hints_monitor, 'changed', self, 'on_hints_file_changed')

            self.refresh_hints()

    def refresh_hints(self):
        self.project.pycore.module_cache.forget_all_data()
        self.project.pycore.hintdb = CompositeHintProvider(self.project)

        if self.hints_filename and os.path.exists(self.hints_filename):
            namespace = {}
            execfile(self.hints_filename, namespace)
            if 'init' in namespace:
                try:
                    namespace['init'](self.project.pycore.hintdb)
                except:
                    import traceback
                    traceback.print_exc()

    def on_hints_file_changed(self, filemonitor, file, other_file, event):
        if event in (gio.FILE_MONITOR_EVENT_CHANGES_DONE_HINT, gio.FILE_MONITOR_EVENT_CREATED):
            self.refresh_hints()

    def close(self):
        if self.hints_monitor:
            self.hints_monitor.cancel()
            self.hints_monitor = None
        self.project.close()

    def __del__(self):
        self.close()

class Plugin(object):
    def __init__(self, editor):
        self.editor = editor
        idle(connect_all, self, view=editor.view)
        idle(self.init_completion)

    def init_completion(self):
        provider = self.completion_provider
        completion = self.editor.view.get_completion()
        completion.add_provider(provider)
        #completion.get_info_window().set_sizing(300, 400, False, False)

    @lazy_property
    def project_manager(self):
        root = getattr(self.editor, 'ropeproject_root', self.editor.project_root)
        try:
            return project_managers[root]
        except KeyError:
            pass

        if os.access(root, os.W_OK):
            kwargs = {}
        else:
            kwargs = dict(ropefolder=None)

        from rope.base.project import Project
        project = Project(root, **kwargs)
        project.snaked_project_root = root

        pm = RopeProjectManager(project)
        project_managers[root] = pm
        return pm

    @lazy_property
    def completion_provider(self):
        import complete
        return complete.RopeCompletionProvider(self)

    def get_rope_resource(self, project, uri=None):
        from rope.base import libutils, exceptions
        uri = uri or self.editor.uri
        try:
            return libutils.path_to_resource(project, uri)
        except exceptions.ResourceNotFoundError:
            return None

    def get_source_and_offset(self):
        offset = self.editor.cursor.get_offset()
        source = self.editor.text

        if not isinstance(source, unicode):
            source = source.decode('utf8')

        return source, offset

    def get_fuzzy_location(self, project, source, offset):
        from rope.base import worder, exceptions

        word_finder = worder.Worder(source, True)
        expression = word_finder.get_primary_at(offset)
        expression = expression.replace('\\\n', ' ').replace('\n', ' ')

        names = expression.split('.')
        pyname = None
        try:
            obj = project.pycore.get_module(names[0])
            for n in names[1:]:
                pyname = obj[n]
                obj = pyname.get_object()
        except (exceptions.ModuleNotFoundError, exceptions.AttributeNotFoundError):
            return None, None

        if not pyname:
            try:
                resource = obj._get_init_dot_py()
            except AttributeError:
                resource = obj.get_resource()

            if not resource:
                return None, None
            else:
                return resource, 1
        else:
            resource, line = pyname.get_definition_location()
            if hasattr(resource, 'resource'):
                resource = resource.resource

        return resource, line

    def goto_definition(self):
        project = self.project_manager.project

        project.validate()

        current_resource = self.get_rope_resource(project)

        from rope.contrib import codeassist

        source, offset = self.get_source_and_offset()
        try:
            resource, line = codeassist.get_definition_location(
                project, source, offset,
                resource=current_resource, maxfixes=3)
        except Exception, e:
            import traceback
            traceback.print_exc()
            self.editor.message(str(e), 5000)
            return

        if resource is None and line is None:
            resource, line = self.get_fuzzy_location(project, source, offset)

        if resource and resource.real_path == current_resource.real_path:
            resource = None

        if resource:
            uri = resource.real_path
            editor = self.editor.open_file(uri, line - 1)
            editor.ropeproject_root = project.snaked_project_root
        else:
            if line:
                self.editor.add_spot()
                self.editor.goto_line(line)
            else:
                self.editor.message("Unknown definition")

    @connect_external('view', 'key-press-event')
    def on_textview_key_press_event(self, sender, event):
        if event.keyval != gtk.keysyms.Return:
            return False

        cursor = self.editor.cursor
        line_start = cursor.copy()
        line_start.set_line(line_start.get_line())

        text = line_start.get_text(cursor).strip()
        if text and text[-1] == ':':
            end = line_start.copy()
            end.forward_word_end()
            end.backward_word_start()
            ws = line_start.get_text(end)

            if self.editor.view.get_insert_spaces_instead_of_tabs():
                tab = u' ' * self.editor.view.get_tab_width()
            else:
                tab = u'\t'

            self.editor.buffer.begin_user_action()
            self.editor.buffer.insert(cursor, u'\n' + ws + tab)
            self.editor.buffer.end_user_action()

            idle(self.editor.view.scroll_mark_onscreen, self.editor.buffer.get_insert())

            return True

        return False

    @connect_external('view', 'backspace')
    def on_textview_backspace(self, *args):
        cursor = self.editor.cursor

        if cursor.starts_line():
            return False

        start = cursor.copy()
        start.set_line(start.get_line())

        text = start.get_text(cursor)

        if text.strip():
            return False

        delete_from = cursor.copy()
        if text[-1] == u'\t':
            delete_from.backward_char()
        else:
            delete_from.backward_chars(self.editor.view.get_tab_width() - 1)

        if delete_from.get_line() != start.get_line():
            delete_from = start

        if delete_from.equal(start):
            delete_from.forward_char()

        self.editor.buffer.begin_user_action()
        self.editor.buffer.delete(delete_from, cursor)
        self.editor.buffer.end_user_action()

        return True

    def show_calltips(self):
        project = self.project_manager.project
        project.validate()

        current_resource = self.get_rope_resource(project)

        from rope.contrib import codeassist
        from snaked.util.pairs_parser import get_brackets

        source, offset = self.get_source_and_offset()

        brackets = get_brackets(source, offset)
        if brackets:
            br, spos, epos = brackets
            if br == '(':
                offset = spos - 1

        try:
            doc = codeassist.get_doc(project, source, offset, resource=current_resource, maxfixes=3)
        except Exception, e:
            import traceback
            traceback.print_exc()
            self.editor.message(str(e), 5000)
            return

        if doc:
            self.editor.message(doc.strip(), 20000)
        else:
            self.editor.message('Info not found')


import re

def get_pattern(what):
    parts = what.split('/')

    if len(parts) <= 1:
        return None

    rexp = r'(/|^)' + r'[^/]+/'.join(map(re.escape, parts))
    return re.compile(rexp)

def name_start_match(what):
    def inner(name, path):
        return name.startswith(what)

    return inner

def name_match(what):
    def inner(name, path):
        return what in name

    return inner

def path_match(what):
    def inner(name, path):
        return what in path

    return inner

def fuzzy_match(what):
    pattern = get_pattern(what)
    def inner(name, path):
        if pattern:
            return pattern.search(path) is not None
        else:
            return False

    return inner

def dir_is_good(name, path):
    if name.startswith('.'):
        return False

    return True

def file_is_good(name, path):
    if any(map(name.endswith, ('.pyc', '.pyo', '.png'))):
        return False

    return True

def search(root, top, match, already_matched, tick):
    from os import listdir
    from os.path import join, isdir

    tick()

    dirs_to_visit = []
    for name in listdir(join(root, top)):
        tick()
        fullpath = join(root, top, name)
        path = join(top, name)

        if isdir(fullpath) and dir_is_good(name, path):
            dirs_to_visit.append(path)
        elif (name, top) not in already_matched and match(name, path) and file_is_good(name, path):
            yield name, top

    for path in dirs_to_visit:
        for p in search(root, path, match, already_matched, tick):
            yield p



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Edit and Select'
desc = 'Various edit shortcuts'

import gtk

def init(manager):
    manager.add_shortcut('delete-line', '<ctrl>d', 'Edit', 'Deletes current line', delete_line)
    manager.add_shortcut('smart-select', '<alt>w', 'Selection', 'Smart anything selection', smart_select)
    manager.add_shortcut('show_offset', '<ctrl><alt>o', 'Edit',
        'Show cursor offset and column', show_offset)
    manager.add_shortcut('wrap-text', '<alt>f', 'Edit', 'Wrap text on right margin width', wrap_text)

def delete_line(editor):
    from util import get_line_bounds
    bounds = get_line_bounds(editor.cursor)
    clipboard = editor.view.get_clipboard(gtk.gdk.SELECTION_CLIPBOARD)
    editor.buffer.select_range(*bounds)
    editor.buffer.copy_clipboard(clipboard)

    editor.buffer.begin_user_action()
    editor.buffer.delete(*bounds)
    editor.buffer.end_user_action()

def smart_select(editor):
    from smart_select import get_smart_select
    start, end = get_smart_select(editor)
    editor.buffer.select_range(end, start)

def show_offset(editor):
    editor.message('offset: %d\ncolumn: %d' % (
        editor.cursor.get_offset(), editor.cursor.get_line_offset()), 3000)

def wrap_text(editor):
    buf = editor.buffer
    if not buf.get_has_selection():
        editor.message('Select text block to wrap')
        return

    import textwrap

    start, end = buf.get_selection_bounds()
    start.order(end)
    if end.starts_line():
        end.backward_visible_cursor_position()

    text = buf.get_text(start, end).decode('utf-8')
    text = textwrap.fill(text, width=editor.view.get_right_margin_position())

    buf.begin_user_action()
    buf.place_cursor(end)
    buf.delete(start, end)
    buf.insert_at_cursor(text)
    buf.end_user_action()


import weakref

import gtk

import snaked.core.shortcuts
import snaked.core.manager

class WindowedEditorManager(snaked.core.manager.EditorManager):
    def __init__(self):
        super(WindowedEditorManager, self).__init__()
        self.windows = weakref.WeakKeyDictionary()
        self.activators = weakref.WeakKeyDictionary()

    def on_window_destroy(self, window, editor):
        editor().editor_closed.emit()

    def manage_editor(self, editor):
        window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        window.set_default_size(800, 550)
        window.add(editor.widget)

        editor.view.grab_focus()

        weak_editor = weakref.ref(editor)
        def ctx():
            return (weak_editor(), )

        activator = snaked.core.shortcuts.ContextShortcutActivator(window, ctx)
        self.windows[editor] = window
        self.activators[editor] = activator

        window.connect('destroy', self.on_window_destroy, weak_editor)
        window.show_all()

    def focus_editor(self, editor):
        self.windows[editor].present()

    def set_editor_title(self, editor, title):
        self.windows[editor].set_title(title)

    def close_editor(self, editor):
        self.windows[editor].destroy()

    def set_editor_shortcuts(self, editor):
        activator = self.activators[editor]
        self.plugin_manager.bind_shortcuts(activator, editor)

        activator.bind_to_name('quit', self.quit)
        activator.bind_to_name('close-window', self.close_editor)
        activator.bind_to_name('save', self.save)
        activator.bind_to_name('new-file', self.new_file_action)

        activator.bind_to_name('show-preferences', self.show_preferences)

        activator.bind('Escape', self.process_escape)

    def save(self, editor):
        editor.save()

    def set_transient_for(self, editor, window):
        try:
            window.set_transient_for(self.windows[editor])
        except KeyError:
            pass



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Save positions'
desc = 'Remembers last edit position for every edited file'

prefs = None

def get_prefs():
    global prefs
    if not prefs:
        from snaked.core.prefs import KVSettings
        prefs = KVSettings('positions.db')
    
    return prefs

def editor_created(editor):
    editor.connect('get-file-position', on_editor_get_file_position)

def on_editor_get_file_position(editor):
    prefs = get_prefs()
    if editor.uri in prefs:
        return int(prefs[editor.uri])
    
    return -1

def editor_closed(editor):
    prefs = get_prefs()
    prefs[editor.uri] = str(editor.cursor.get_line())

def quit():
    global prefs
    if prefs:
        del prefs



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Hash comment'
desc = '(Un)Comments line or selection with hashes'

import gtk
import re

langs = ['python', 'sh', 'ruby', 'perl']

def init(manager):
    manager.add_shortcut('comment-code', '<ctrl>slash', 'Edit',
        'Comments/uncomments current line or selection', comment)

#Action
def comment(editor):
    r = get_bounds(editor)
    traversor = make_line_traversor(editor.buffer, r)

    if all(line_is_hashed(l) for l in traversor()):
        uncomment_range(editor, traversor)
    else:
        comment_range(editor, traversor)

def make_line_traversor(buffer, r):
    start, end = r
    start, stop = start.get_line(), end.get_line() + 1
    def inner():
        for i in xrange(start, stop):
            yield buffer.get_iter_at_line(i)

    return inner

def get_bounds(editor):
    if editor.buffer.get_has_selection():
        start, end = editor.buffer.get_selection_bounds()
        if start.ends_line():
            start.set_line(start.get_line() + 1)

        if end.starts_line():
            end.set_line(end.get_line() - 1)

        return start, end
    else:
        cursor = editor.cursor
        return cursor, cursor.copy()

def line_is_hashed(iter):
    text = get_line_text(iter).strip()
    return not text or text[0] == u'#'

def get_line_text(iter):
    if not iter.starts_line():
        iter = iter.copy()
        iter.set_line(iter.get_line())

    end = iter.copy()
    if not end.ends_line():
        end.forward_to_line_end()

    return iter.get_text(end)

def line_is_empty(iter):
    return get_line_text(iter).strip() == u''

def comment_range(editor, traversor):
    ws_match = re.compile(r'[ \t]*')
    min_indent = 1000
    min_indent_text = ''
    for iter in traversor():
        if not line_is_empty(iter):
            match = ws_match.match(get_line_text(iter))
            if match and len(match.group()) < min_indent:
                min_indent = len(match.group())
                min_indent_text = match.group()

    if min_indent == 1000:
        min_indent = 0

    editor.buffer.begin_user_action()
    for iter in traversor():
        if line_is_empty(iter):
            line_end = iter.copy()
            line_end.forward_to_line_end()
            if iter.get_line() == line_end.get_line():
                editor.buffer.delete(iter, line_end)
            editor.buffer.insert(iter, min_indent_text)
        else:
            iter.forward_chars(min_indent)

        editor.buffer.insert(iter, u'#')

    editor.buffer.end_user_action()

def uncomment_range(editor, traversor):
    editor.buffer.begin_user_action()
    for iter in traversor():
        if not line_is_empty(iter):
            if get_line_text(iter).strip() == '#':
                line_end = iter.copy()
                line_end.forward_to_line_end()
                editor.buffer.delete(iter, line_end)
            else:
                editor.buffer.delete(*iter.forward_search(u'#', gtk.TEXT_SEARCH_VISIBLE_ONLY))

    editor.buffer.end_user_action()



import os
import shutil
from os.path import join, dirname, realpath, abspath, exists
import re

import gobject
import gtk

from snaked.signals.signals import Handler
from snaked.signals import weak_connect

def idle_callback(callable, args):
    args, kwargs = args
    callable(*args, **kwargs)
    return False

def idle(callable, *args, **kwargs):
    return gobject.idle_add(idle_callback, callable, (args, kwargs))

def save_file(filename, data, encoding):
    tmpfilename = realpath(filename) + '.bak'

    try:
        f = open(tmpfilename, 'w')
    except IOError:
        dname = dirname(tmpfilename)
        if not exists(dname):
            os.makedirs(dname, mode=0755)
            f = open(tmpfilename, 'w')
        else:
            raise

    f.write(data.encode(encoding))
    f.close()

    if exists(filename):
        shutil.copymode(filename, tmpfilename)

    os.rename(tmpfilename, filename)

def connect(sender, signal, obj, attr, idle=False, after=False):
    return Handler(weak_connect(
        sender, signal, obj, attr, idle=idle, after=after), sender)

def join_to_file_dir(filename, *args):
    return join(dirname(filename), *args)

def get_project_root(filename):
    path = dirname(abspath(filename))
    magic_pathes = ['.git', '.hg', '.bzr', '.ropeproject', '.snaked_project']

    while True:
        if any(exists(join(path, p)) for p in magic_pathes):
            return path

        parent = dirname(path)
        if parent == path:
            return None

        path = parent

def open_mime(filename):
    import subprocess
    subprocess.Popen(['/usr/bin/env', 'xdg-open', filename]).poll()


def refresh_gui():
    while gtk.events_pending():
        gtk.main_iteration_do(block=False)

def single_ref(func):
    real_name = '__' + func.__name__
    holder_name = func.__name__ + '_holder'

    def inner(self):
        try:
            return getattr(self, real_name)
        except AttributeError:
            pass

        cls = self.__class__
        if not hasattr(cls, holder_name) or not getattr(cls, holder_name)():
            import weakref
            var = func(self)
            setattr(cls, holder_name, weakref.ref(var))

        setattr(self, real_name, getattr(cls, holder_name)())
        return getattr(self, real_name)

    return property(inner)

def lazy_property(func):
    real_name = '__' + func.__name__

    def inner(self):
        try:
            return getattr(self, real_name)
        except AttributeError:
            pass

        var = func(self)
        setattr(self, real_name, var)
        return var

    return property(inner)

def set_activate_the_one_item(entry, treeview):
    def activate(*args):
        if len(treeview.get_model()) == 1:
            treeview.set_cursor((0,))
            treeview.row_activated((0,), treeview.get_column(0))

    entry.connect('activate', activate)

def create_lang_matchers_from_file(project_root, filename):
    result = []
    for l in open(filename):
        l = l.strip()
        if l.startswith('#'):
            continue

        try:
            expr, contexts = l.split(':')
        except ValueError:
            continue

        contexts = [c.strip() for c in contexts.split(',')]
        if not contexts:
            continue

        if not l.startswith('/'):
            expr = '/*/' + expr

        matcher = re.compile(project_root + re.escape(expr).replace('\*', '.*') + '$')
        result.append((matcher, contexts))

    return result


class BuilderAware(object):
    def __init__(self, glade_file):
        self.gtk_builder = gtk.Builder()
        self.gtk_builder.add_from_file(glade_file)
        self.gtk_builder.connect_signals(self)

    def __getattr__(self, name):
        obj = self.gtk_builder.get_object(name)
        if not obj:
            raise AttributeError('Builder have no %s object' % name)

        setattr(self, name, obj)
        return obj


class LangGuesser(object):
    def __init__(self, matchers):
        self.matchers = matchers

    def guess(self, filename):
        for m, ctx in self.matchers:
            if m.match(filename):
                return ctx

        return []


class RegexObject:
    def __init__(self):
        self.flags = 0
        self.groups = 0
        self.groupindex = {}
        self.pattern = ""

    def match(string, pos=None, endpos=None):
        """If zero or more characters at the beginning of *string* match this regular
        expression, return a corresponding :class:`MatchObject` instance.  Return
        ``None`` if the string does not match the pattern; note that this is different
        from a zero-length match.

        .. note::

           If you want to locate a match anywhere in *string*, use
           :meth:`~RegexObject.search` instead.

        The optional second parameter *pos* gives an index in the string where the
        search is to start; it defaults to ``0``.  This is not completely equivalent to
        slicing the string; the ``'^'`` pattern character matches at the real beginning
        of the string and at positions just after a newline, but not necessarily at the
        index where the search is to start.

        The optional parameter *endpos* limits how far the string will be searched; it
        will be as if the string is *endpos* characters long, so only the characters
        from *pos* to ``endpos - 1`` will be searched for a match.  If *endpos* is less
        than *pos*, no match will be found, otherwise, if *rx* is a compiled regular
        expression object, ``rx.match(string, 0, 50)`` is equivalent to
        ``rx.match(string[:50], 0)``.

           >>> pattern = re.compile("o")
           >>> pattern.match("dog")      # No match as "o" is not at the start of "dog."
           >>> pattern.match("dog", 1)   # Match as "o" is the 2nd character of "dog".
           <_sre.SRE_Match object at ...>

        """
        return MatchObject()

    def search(string, pos=None, endpos=None):
        """Scan through *string* looking for a location where this regular expression
        produces a match, and return a corresponding :class:`MatchObject` instance.
        Return ``None`` if no position in the string matches the pattern; note that this
        is different from finding a zero-length match at some point in the string.

        The optional *pos* and *endpos* parameters have the same meaning as for the
        :meth:`~RegexObject.match` method.
        """

        return MatchObject()

    def split(string, maxsplit=0):
        """Split *string* by the occurrences of *pattern*.

        If capturing parentheses are
        used in *pattern*, then the text of all groups in the pattern are also returned
        as part of the resulting list. If *maxsplit* is nonzero, at most *maxsplit*
        splits occur, and the remainder of the string is returned as the final element
        of the list.  (Incompatibility note: in the original Python 1.5 release,
        *maxsplit* was ignored.  This has been fixed in later releases.)

            >>> re.split('\W+', 'Words, words, words.')
            ['Words', 'words', 'words', '']
            >>> re.split('(\W+)', 'Words, words, words.')
            ['Words', ', ', 'words', ', ', 'words', '.', '']
            >>> re.split('\W+', 'Words, words, words.', 1)
            ['Words', 'words, words.']
        """

    def findall(string, pos=None, endpos=None):
        """Return all non-overlapping matches of *pattern* in *string*, as a list of strings.

        The *string* is scanned left-to-right, and matches are returned in
        the order found.  If one or more groups are present in the pattern, return a
        list of groups; this will be a list of tuples if the pattern has more than
        one group.  Empty matches are included in the result unless they touch the
        beginning of another match.
        """

    def finditer(string, pos=None, endpos=None):
        """Return an :term:`iterator` yielding :class:`MatchObject` instances over all
        non-overlapping matches for the RE *pattern* in *string*.

        The *string* is
        scanned left-to-right, and matches are returned in the order found.  Empty
        matches are included in the result unless they touch the beginning of another
        match.
        """

    def sub(repl, string, count=0):
        r"""Return the string obtained by replacing the leftmost non-overlapping occurrences
        of *pattern* in *string* by the replacement *repl*.  If the pattern isn't found,
        *string* is returned unchanged.  *repl* can be a string or a function; if it is
        a string, any backslash escapes in it are processed.  That is, ``\n`` is
        converted to a single newline character, ``\r`` is converted to a linefeed, and
        so forth.  Unknown escapes such as ``\j`` are left alone.  Backreferences, such
        as ``\6``, are replaced with the substring matched by group 6 in the pattern.
        For example:

          >>> re.sub(r'def\s+([a-zA-Z_][a-zA-Z_0-9]*)\s*\(\s*\):',
          ...        r'static PyObject*\npy_\1(void)\n{',
          ...        'def myfunc():')
          'static PyObject*\npy_myfunc(void)\n{'

        If *repl* is a function, it is called for every non-overlapping occurrence of
        *pattern*.  The function takes a single match object argument, and returns the
        replacement string.  For example:

          >>> def dashrepl(matchobj):
          ...     if matchobj.group(0) == '-': return ' '
          ...     else: return '-'
          >>> re.sub('-{1,2}', dashrepl, 'pro----gram-files')
          'pro--gram files'

        The pattern may be a string or an RE object; if you need to specify regular
        expression flags, you must use a RE object, or use embedded modifiers in a
        pattern; for example, ``sub("(?i)b+", "x", "bbbb BBBB")`` returns ``'x x'``.

        The optional argument *count* is the maximum number of pattern occurrences to be
        replaced; *count* must be a non-negative integer.  If omitted or zero, all
        occurrences will be replaced. Empty matches for the pattern are replaced only
        when not adjacent to a previous match, so ``sub('x*', '-', 'abc')`` returns
        ``'-a-b-c-'``.

        In addition to character escapes and backreferences as described above,
        ``\g<name>`` will use the substring matched by the group named ``name``, as
        defined by the ``(?P<name>...)`` syntax. ``\g<number>`` uses the corresponding
        group number; ``\g<2>`` is therefore equivalent to ``\2``, but isn't ambiguous
        in a replacement such as ``\g<2>0``.  ``\20`` would be interpreted as a
        reference to group 20, not a reference to group 2 followed by the literal
        character ``'0'``.  The backreference ``\g<0>`` substitutes in the entire
        substring matched by the RE.
        """

    def subn(repl, string, count=0):
        """Perform the same operation as :func:`sub`, but return a tuple ``(new_string,
        number_of_subs_made)``.
        """

class MatchObject:
    def __init__(self):
        self.pos = 0
        self.endpos = 0
        self.lastindex = 0
        self.lastgroup = ""
        self.re = RegexObject()
        self.string = ""

    def expand(template):
        """Return the string obtained by doing backslash substitution on the template
        string *template*, as done by the :meth:`~RegexObject.sub` method.  Escapes
        such as ``\n`` are converted to the appropriate characters, and numeric
        backreferences (``\1``, ``\2``) and named backreferences (``\g<1>``,
        ``\g<name>``) are replaced by the contents of the corresponding group.
        """

    def group(*groups):
        """
        Returns one or more subgroups of the match.  If there is a single argument, the
        result is a single string; if there are multiple arguments, the result is a
        tuple with one item per argument. Without arguments, *group1* defaults to zero
        (the whole match is returned). If a *groupN* argument is zero, the corresponding
        return value is the entire matching string; if it is in the inclusive range
        [1..99], it is the string matching the corresponding parenthesized group.  If a
        group number is negative or larger than the number of groups defined in the
        pattern, an :exc:`IndexError` exception is raised. If a group is contained in a
        part of the pattern that did not match, the corresponding result is ``None``.
        If a group is contained in a part of the pattern that matched multiple times,
        the last match is returned.

          >>> m = re.match(r"(\w+) (\w+)", "Isaac Newton, physicist")
          >>> m.group(0)       # The entire match
          'Isaac Newton'
          >>> m.group(1)       # The first parenthesized subgroup.
          'Isaac'
          >>> m.group(2)       # The second parenthesized subgroup.
          'Newton'
          >>> m.group(1, 2)    # Multiple arguments give us a tuple.
          ('Isaac', 'Newton')

        If the regular expression uses the ``(?P<name>...)`` syntax, the *groupN*
        arguments may also be strings identifying groups by their group name.  If a
        string argument is not used as a group name in the pattern, an :exc:`IndexError`
        exception is raised.

        A moderately complicated example:

          >>> m = re.match(r"(?P<first_name>\w+) (?P<last_name>\w+)", "Malcom Reynolds")
          >>> m.group('first_name')
          'Malcom'
          >>> m.group('last_name')
          'Reynolds'

        Named groups can also be referred to by their index:

          >>> m.group(1)
          'Malcom'
          >>> m.group(2)
          'Reynolds'

        If a group matches multiple times, only the last match is accessible:

          >>> m = re.match(r"(..)+", "a1b2c3")  # Matches 3 times.
          >>> m.group(1)                        # Returns only the last match.
          'c3'
        """

    def groups(default=None):
        """
        Return a tuple containing all the subgroups of the match, from 1 up to however
        many groups are in the pattern.  The *default* argument is used for groups that
        did not participate in the match; it defaults to ``None``.  (Incompatibility
        note: in the original Python 1.5 release, if the tuple was one element long, a
        string would be returned instead.  In later versions (from 1.5.1 on), a
        singleton tuple is returned in such cases.)

        For example:

          >>> m = re.match(r"(\d+)\.(\d+)", "24.1632")
          >>> m.groups()
          ('24', '1632')

        If we make the decimal place and everything after it optional, not all groups
        might participate in the match.  These groups will default to ``None`` unless
        the *default* argument is given:

          >>> m = re.match(r"(\d+)\.?(\d+)?", "24")
          >>> m.groups()      # Second group defaults to None.
          ('24', None)
          >>> m.groups('0')   # Now, the second group defaults to '0'.
          ('24', '0')
        """

    def groupdict(default):
        """
        Return a dictionary containing all the *named* subgroups of the match, keyed by
        the subgroup name.  The *default* argument is used for groups that did not
        participate in the match; it defaults to ``None``.  For example:

          >>> m = re.match(r"(?P<first_name>\w+) (?P<last_name>\w+)", "Malcom Reynolds")
          >>> m.groupdict()
          {'first_name': 'Malcom', 'last_name': 'Reynolds'}
        """

        return {}

    def start(group=None):
        """Return the indices of the start and end of the substring matched by *group*;
        *group* defaults to zero (meaning the whole matched substring). Return ``-1`` if
        *group* exists but did not contribute to the match.  For a match object *m*, and
        a group *g* that did contribute to the match, the substring matched by group *g*
        (equivalent to ``m.group(g)``) is ::

          m.string[m.start(g):m.end(g)]

        Note that ``m.start(group)`` will equal ``m.end(group)`` if *group* matched a
        null string.  For example, after ``m = re.search('b(c?)', 'cba')``,
        ``m.start(0)`` is 1, ``m.end(0)`` is 2, ``m.start(1)`` and ``m.end(1)`` are both
        2, and ``m.start(2)`` raises an :exc:`IndexError` exception.

        An example that will remove *remove_this* from email addresses:

          >>> email = "tony@tiremove_thisger.net"
          >>> m = re.search("remove_this", email)
          >>> email[:m.start()] + email[m.end():]
          'tony@tiger.net'
        """

    def end(group=None):
        """see :method:`start`"""

    def span(group=None):
        """For :class:`MatchObject` *m*, return the 2-tuple ``(m.start(group),
        m.end(group))``. Note that if *group* did not contribute to the match, this is
        ``(-1, -1)``.  *group* defaults to zero, the entire match.
        """


import re

class Snippet(object):
    def __init__(self, snippet, variant):
        self.snippet = snippet
        self.variant = variant
        self.comment = ''
        self.body = []

        if variant:
            self.label = snippet + ' ' + variant
        else:
            self.label = snippet

    def get_body_and_offsets(self, indent=u'', expand_tabs=False, tab_width=4):
        tab_offsets = {}
        insert_offsets = {}
        replaces = {}
        matcher = re.compile(ur'\$\{(\d+)(:(.*?))?\}')

        body = (u'\n' + indent).join(
            s.expandtabs(tab_width) if expand_tabs else s for s in self.body)

        for m in matcher.finditer(body):
            if m.group(3):
                replaces[int(m.group(1))] = m.group(3)

        delta = [0]
        def replace_stops(match):
            idx = int(match.group(1))
            replace = replaces.get(idx, u'')

            start = delta[0] + match.start()
            delta[0] += len(replace) - match.end() + match.start()
            end = delta[0] + match.end()

            tab_offsets[idx] = start, end

            return replace

        def replace_inserts(match):
            idx = int(match.group(1))
            replace = replaces.get(idx, u'')

            start = delta[0] + match.start()
            dt = len(replace) - match.end() + match.start()
            delta[0] += dt
            end = delta[0] + match.end()

            for k, (s, e) in tab_offsets.iteritems():
                if s >= start: s += dt
                if e >= start: e += dt
                tab_offsets[k] = s, e

            insert_offsets[idx] = start, end

            return replace

        body = matcher.sub(replace_stops, body)

        delta[0] = 0
        body = re.sub(ur'\$(\d+)', replace_inserts, body)

        return body, tab_offsets, insert_offsets


def parse_snippets_from(filename):
    pl = ''
    csnippet = None
    snippets = {}
    for l in open(filename).read().decode('utf-8').splitlines():
        if l.startswith('snippet'):
            tag_and_variant = l.split(None, 2)[1:]
            if len(tag_and_variant) == 2:
                tag, variant = tag_and_variant
            else:
                tag, variant = tag_and_variant[0], None
            csnippet = Snippet(tag, variant)
            snippets[csnippet.label] = csnippet
            if pl.startswith('#'):
                csnippet.comment = pl[1:].strip()
        elif l.startswith('\t') and csnippet:
            csnippet.body.append(l[1:])

        pl = l

    return snippets


from optparse import OptionParser
import os

def select_session():
    import gtk
    from snaked.util import join_to_file_dir
    from snaked.core.prefs import get_settings_path
    
    builder = gtk.Builder()
    builder.add_from_file(join_to_file_dir(__file__, 'gui', 'select_session.glade'))
    dialog = builder.get_object('dialog')
    dialog.vbox.remove(dialog.action_area)
    dialog.set_default_response(1)
    
    sessions_view = builder.get_object('sessions_view')
    sessions = builder.get_object('sessions')
    
    for p in os.listdir(get_settings_path('')):
        if p.endswith('.session'):
            sessions.append((p.rpartition('.')[0],))

    def row_activated(view, path, *args):
        dialog.response(path[0])
        
    sessions_view.connect('row-activated', row_activated)
    
    response = dialog.run()
    result = sessions[response][0] if response >= 0 else None
    dialog.destroy()    
    return result

def get_manager():
    parser = OptionParser()
    parser.add_option('-s', '--session', dest='session', help="Open snaked with specified session")
    parser.add_option('-w', '--windowed', action="store_true", 
        dest='windowed', default=False, help="Open separate editor window instead tab")
    parser.add_option('', '--select-session', action="store_true", dest='select_session',
        help="Opens dialog to select session", default=False)
    parser.add_option('', '--hide-tabs', action="store_false", dest='show_tabs',
        help="Hides tabs in tabbed interface", default=True)

    options, args = parser.parse_args()

    import gobject
    gobject.threads_init()
    
    if options.windowed:
        from .windowed import WindowedEditorManager as WM
        manager = WM()
    else:
        from .tabbed import TabbedEditorManager as WM
        manager = WM(options.show_tabs)

    if options.select_session:
        options.session = select_session()

    opened_files = []
    
    session_files = []
    active_file = None
    if options.session:
        settings = manager.get_session_settings(options.session)
        session_files = settings.get('files', [])
        active_file = settings.get('active_file', None)
        manager.session = options.session
    
    editor_to_focus = None
    for f in session_files + args:
        f = os.path.abspath(f)
        if f not in opened_files and (not os.path.exists(f) or os.path.isfile(f)):    
            e = manager.open(f)
            if f == active_file:
                editor_to_focus = e
            opened_files.append(f)

    if not manager.editors:
        import snaked.core.quick_open
        snaked.core.quick_open.activate(manager.get_fake_editor())

    if editor_to_focus and active_file != opened_files[-1]:
        manager.focus_editor(editor_to_focus)
        
    return manager
            
def run():
    manager = get_manager()

    import gtk
    
    try:    
        gtk.main()
    except KeyboardInterrupt:
        manager.quit(None)



from setuptools import setup, find_packages

setup(
    name     = 'snaked',
    version  = '0.4.3',
    author   = 'Anton Bobrov',
    author_email = 'bobrov@vl.ru',
    description = 'Very light and minimalist editor inspired by Scribes',
    long_description = open('README.rst').read(),
    zip_safe   = False,
    packages = find_packages(),
    install_requires = ['chardet'],
    include_package_data = True,
    namespace_packages = ['snaked', 'snaked.plugins'],
    entry_points = {
        'gui_scripts': [
            'snaked = snaked.core.run:run',
        ]
    },
    url = 'http://github.com/baverman/snaked',
    classifiers = [
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
        "Environment :: X11 Applications :: GTK",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Topic :: Text Editors"
    ],
)



import re

def get_line_bounds(cursor):
    end = cursor.copy()
    start = end.copy()
    start.set_line(end.get_line())

    very_end = end.copy()
    very_end.forward_to_end()

    if end.get_line() == very_end.get_line():
        start.backward_char()
        return start, very_end

    if not end.ends_line():
        end.forward_to_line_end()
        
    end.forward_char()
    
    return start, end    
    
def cursor_on_start_or_end_whitespace(cursor):
    if cursor.starts_line() or cursor.ends_line():
        return True
    
    start, end = get_line_bounds(cursor)
    starttext = start.get_text(cursor)
    endtext = cursor.get_text(end)

    if starttext.strip() == u'' or endtext.strip() == u'':
        return True
             
    return False

match_ws = re.compile(u'(?u)^[ \t]*')
def get_whitespace(start):
    if start.is_end():
        return u''
        
    match = match_ws.search(line_text(start))
    if match:
        return match.group(0)
    else:
        return u''

def line_text(iter):
    if not iter.starts_line():
        iter = iter.copy()
        iter.set_line(iter.get_line())

    end = iter.copy()
    if not end.ends_line():
        end.forward_to_line_end()
    
    return iter.get_text(end)
        
def line_is_empty(iter):
    return iter.is_end() or line_text(iter).strip() == u''

def iter_lines(from_iter, delta):
    line_count = from_iter.get_buffer().get_line_count()
    iter = from_iter.copy()
    while True:
        newline = iter.get_line() + delta
        if newline < 0 or newline > line_count - 1:
            return
        
        olditer = iter.copy()
        iter.set_line(iter.get_line() + delta)
        
        yield olditer, iter

def get_next_not_empty_line(from_iter, delta):
    for p, n in iter_lines(from_iter, delta):
        if not line_is_empty(n):
            return n
            
    return None

def next_line(iter):
    result = iter.copy()
    result.set_line(result.get_line() + 1)
    return result

def prev_line(iter):
    result = iter.copy()
    result.set_line(result.get_line() - 1)
    return result


import weakref

import re

from snaked.util import BuilderAware, join_to_file_dir, idle
from snaked.core import shortcuts
from snaked.core.prefs import CompositePreferences, load_json_settings, save_json_settings

default_prefs = {
    'name': '',
    'langs': '',
    'command': '',
    'stdin': 'none',
    'stdout': 'show-feedback',
}

remove_tags = re.compile(r'<[^<]*?/?>')

class PreferencesDialog(BuilderAware):
    def __init__(self):
        BuilderAware.__init__(self, join_to_file_dir(__file__, 'prefs.glade'))
        self.activator = shortcuts.ShortcutActivator(self.window)
        self.activator.bind('Escape', self.hide)
        self.activator.bind('Delete', self.delete_tool)

        self.dirty_settings = {}
        self.tool_settings = load_json_settings('external-tools.conf', {})

        for name in sorted(self.tool_settings, reverse=True):
            self.tools.prepend((name,))

        for n in ('name', 'langs', 'command'):
            getattr(self, n).connect('changed', self.entry_changed, n)

        for n in ('stdin', 'stdout'):
            getattr(self, n+'_cb').connect('changed', self.cb_changed, n)

    def get_name(self, name):
        return remove_tags.sub('', name).strip().replace('_', '')

    def show(self, editor):
        self.editor = weakref.ref(editor)
        editor.request_transient_for.emit(self.window)
        self.select_tool('New tool')
        self.tools_view.columns_autosize()
        self.window.present()

    def save_settings(self):
        settings = {}
        for v in self.tool_settings.values():
            settings[self.get_name(v['name'])] = v
        save_json_settings('external-tools.conf', settings)

    def hide(self):
        self.save_settings()
        self.editor().message('External tool settings saved')
        idle(self.window.destroy)

    @property
    def current_tool(self):
        (model, iter) = self.tools_view.get_selection().get_selected()
        if iter:
            return self.tools.get_value(iter, 0)
        else:
            return None

    @property
    def settings(self):
        name = self.current_tool
        if name is None:
            self.vbox1.hide()
            self.editor().message('You need to select any tool', 2000)
            return {}
        elif self.isnew(name):
            return self.dirty_settings
        else:
            return self.tool_settings.setdefault(name, {})

    def isnew(self, tool):
        return tool == 'New tool'

    def entry_changed(self, entry, name):
        self.settings[name] = entry.get_text()

    def cb_changed(self, cb, name):
        self.settings[name] = cb.get_model().get_value(cb.get_active_iter(), 0)

    def select_tool(self, name):
        for i, (n,) in enumerate(self.tools):
            if n == name:
                self.tools_view.set_cursor((i,))
                return

        self.tools_view.set_cursor((len(self.tools)-1,))

    def set_cb(self, cb, value):
        for i, (id, _) in enumerate(cb.get_model()):
            if id == value:
                cb.set_active(i)
                return

        cb.set_active(0)

    def on_tools_view_cursor_changed(self, *args):
        name = self.current_tool
        self.vbox1.show()
        prefs = CompositePreferences(self.settings, default_prefs)

        if self.isnew(name):
            self.add_btn.get_parent().show()
        else:
            self.add_btn.get_parent().hide()

        for n in ('name', 'langs', 'command'):
            getattr(self, n).set_text(prefs[n])

        for n in ('stdin', 'stdout'):
            self.set_cb(getattr(self, n+'_cb'), prefs[n])

    def on_tools_view_toggle_cursor_row(self, *args):
        if self.current_tool is None:
            self.vbox1.show()
        else:
            self.vbox1.hide()

    def on_add_btn_clicked(self, *args):
        name = self.get_name(self.dirty_settings.get('name', ''))
        if not name or self.isnew(name):
            self.editor().message('Enter tool name')
            return
        elif name in self.tool_settings:
            self.editor().message('Tool already exists')
            return

        prefs = CompositePreferences(self.dirty_settings,
            self.tool_settings.get('name', {}), default_prefs)

        new_prefs = self.tool_settings.setdefault(name, {})
        for k in default_prefs:
            new_prefs[k] = prefs[k]

        self.tools.insert(len(self.tools)-1, (name,))
        self.select_tool(name)
        idle(self.tools_view.grab_focus)
        self.dirty_settings.clear()

    def on_window_delete_event(self, *args):
        self.hide()

    def delete_tool(self, *args):
        name = self.current_tool
        if name is None:
            return

        if self.isnew(name):
            self.editor().message('Lolwhat?')
            return

        (model, iter) = self.tools_view.get_selection().get_selected()
        path = model.get_path(iter)
        model.remove(iter)
        del self.tool_settings[name]
        self.tools_view.set_cursor(path)


import time

import gtk

from snaked.core.shortcuts import ContextShortcutActivator, register_shortcut
import snaked.core.manager

class TabbedEditorManager(snaked.core.manager.EditorManager):
    def __init__(self, show_tabs=True):
        super(TabbedEditorManager, self).__init__()

        self.last_switch_time = None

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.connect('delete-event', self.on_delete_event)

        self.window.set_property('default-width', 800)
        self.window.set_property('default-height', 500)

        self.activator = ContextShortcutActivator(self.window, self.get_context)

        self.note = gtk.Notebook()
        self.note.set_show_tabs(show_tabs)
        self.note.set_scrollable(True)
        self.note.set_property('tab-hborder', 10)
        self.note.set_property('homogeneous', False)
        self.note.connect_after('switch-page', self.on_switch_page)
        self.note.connect('page_removed', self.on_page_removed)
        self.window.add(self.note)

        register_shortcut('toggle-tabs-visibility', '<alt>F11', 'Window', 'Toggles tabs visibility')
        register_shortcut('next-editor', '<alt>Right', 'Window', 'Switches to next editor')
        register_shortcut('prev-editor', '<alt>Left', 'Window', 'Switches to previous editor')
        register_shortcut('next-editor-alt', '<ctrl>Page_Down', 'Window', 'Switches to next editor')
        register_shortcut('prev-editor-alt', '<ctrl>Page_Up', 'Window', 'Switches to previous editor')
        register_shortcut('fullscreen', 'F11', 'Window', 'Toggles fullscreen mode')

        self.window.show_all()

    def get_context(self):
        widget = self.note.get_nth_page(self.note.get_current_page())
        for e in self.editors:
            if e.widget is widget:
                return (e,)

        return (None,)

    def manage_editor(self, editor):
        label = gtk.Label('Unknown')
        self.note.insert_page(editor.widget, label, -1)
        self.focus_editor(editor)
        editor.view.grab_focus()

    def focus_editor(self, editor):
        idx = self.note.page_num(editor.widget)
        self.note.set_current_page(idx)

    def update_top_level_title(self):
        idx = self.note.get_current_page()
        if idx < 0:
            return

        title = self.note.get_tab_label_text(self.note.get_nth_page(idx))
        if title is not None:
            self.window.set_title(title)

    def set_editor_title(self, editor, title):
        self.note.set_tab_label_text(editor.widget, title)
        if self.note.get_current_page() == self.note.page_num(editor.widget):
            self.update_top_level_title()

    def on_delete_event(self, *args):
        self.quit(None)

    def close_editor(self, editor):
        idx = self.note.page_num(editor.widget)
        self.note.remove_page(idx)
        editor.editor_closed.emit()

    def set_editor_shortcuts(self, editor):
        self.plugin_manager.bind_shortcuts(self.activator, editor)

        if hasattr(self, 'editor_shortcuts_binded'):
            return

        self.editor_shortcuts_binded = True

        self.activator.bind_to_name('quit', self.quit)
        self.activator.bind_to_name('close-window', self.close_editor)
        self.activator.bind_to_name('save', self.save)
        self.activator.bind_to_name('next-editor', self.switch_to, 1)
        self.activator.bind_to_name('prev-editor', self.switch_to, -1)
        self.activator.bind_to_name('next-editor-alt', self.switch_to, 1)
        self.activator.bind_to_name('prev-editor-alt', self.switch_to, -1)
        self.activator.bind_to_name('new-file', self.new_file_action)
        self.activator.bind_to_name('show-preferences', self.show_preferences)
        self.activator.bind_to_name('fullscreen', self.fullscreen, [True])
        self.activator.bind_to_name('toggle-tabs-visibility', self.toggle_tabs)

        self.activator.bind_to_name('place-spot', self.add_spot_with_feedback)
        self.activator.bind_to_name('goto-last-spot', self.goto_last_spot)
        self.activator.bind_to_name('goto-next-spot', self.goto_next_prev_spot, True)
        self.activator.bind_to_name('goto-prev-spot', self.goto_next_prev_spot, False)

        self.activator.bind('Escape', self.process_escape)

    def quit(self, editor):
        self.window.hide()
        super(TabbedEditorManager, self).quit(editor)

    def save(self, editor):
        editor.save()

    def set_transient_for(self, editor, window):
        window.set_transient_for(self.window)

    def on_switch_page(self, *args):
        self.update_top_level_title()

    def on_page_removed(self, note, child, idx):
        for e in self.editors:
            if e.widget is child:
                spot = self.get_last_spot(None, e)
                if spot:
                    note.set_current_page(note.page_num(spot.editor().widget))
                    return

        if idx > 0:
            note.set_current_page(idx - 1)

    def switch_to(self, editor, dir):
        if self.last_switch_time is None or time.time() - self.last_switch_time > 5:
            self.add_spot(editor)

        self.last_switch_time = time.time()

        idx = ( self.note.get_current_page() + dir ) % self.note.get_n_pages()
        self.note.set_current_page(idx)

    def fullscreen(self, editor, state):
        if state[0]:
            self.window.fullscreen()
        else:
            self.window.unfullscreen()

        state[0] = not state[0]

    def toggle_tabs(self, editor):
        self.note.set_show_tabs(not self.note.get_show_tabs())